# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Queries sent to devices of each type to determine whether they are a particular device type."""
import copy
import enum
import functools
import logging
import re
import requests
import subprocess

from gazoo_device import gdm_logger
from gazoo_device.utility import host_utils
from gazoo_device.utility import http_utils
from gazoo_device.utility import usb_utils
from gazoo_device.utility import supported_classes
logger = gdm_logger.get_gdm_logger()

FMT = "<%(asctime)s> GDM-M: %(message)s"

ADB_COMMANDS = {
}

DOCKER_COMMANDS = {
    "PRODUCT_NAME": "docker ps --filter id={} --format {{{{.Names}}}}"
}

SSH_COMMANDS = {
    "UNIFI_PRODUCT_NAME": "mca-cli-op info",
    "DLI_PRODUCT_NAME": "http://{address}/restapi/config/=brand_name/",
    "RPI_PRODUCT_NAME": "cat /proc/device-tree/model",
}


@functools.total_ordering
class QueryEnum(enum.Enum):
    """Allows comparison of enum properties for sorting purposes."""

    def __lt__(self, other):
        return self.name < other.name  # pylint: disable=comparison-with-callable


class ADB_QUERY(QueryEnum):
    pass


class DOCKER_QUERY(QueryEnum):
    product_name = "product_name"


class GENERIC_QUERY(QueryEnum):
    always_true = "always_true"


class SERIAL_QUERY(QueryEnum):
    product_name = "usb info product_name"


class SSH_QUERY(QueryEnum):
    """Query names for detection for SshComms Devices."""
    is_dli = "is_dli_power_switch"
    is_rpi = "is_raspberry_pi"
    is_unifi = "is_unifi_switch"


def docker_product_name_query(address, detect_logger):
    """Gets product name from docker device.

    Args:
        address (str): communication_address
        detect_logger (Logger): logs device interactions.
    Returns:
        str: product_name or empty
    """
    try:
        name = subprocess.check_output(DOCKER_COMMANDS["PRODUCT_NAME"].format(address).split())
        name = name.decode()
    except subprocess.CalledProcessError as err:
        detect_logger.info("docker_product_name_query failure: " + repr(err))
        return ""
    detect_logger.info("docker_product_name_query response: {}".format(name))
    return name


def always_true_query(address, detect_logger):    # pylint: disable=unused-argument
    """Used when there is just one type of device for a communication type."""
    detect_logger.info("always_true_query response: True")
    return True


def is_dli_query(address, detect_logger):
    """Determines if address belongs to dli power switch.

    Args:
        address (str): ip_address
        detect_logger (Logger): logs device interactions.
    Returns:
        bool: whether or not Power Switch in response
    """
    try:
        response = http_utils.send_http_get(
            SSH_COMMANDS["DLI_PRODUCT_NAME"].format(address=address),
            auth=requests.auth.HTTPDigestAuth("admin", "1234"),
            headers={"Accept": "application/json"},
            valid_return_codes=[200, 206, 207],
            timeout=1)
        name = response.text
    except RuntimeError as err:
        detect_logger.info("is_dli_query failure: " + repr(err))
        return False
    detect_logger.info("is_dli_query response: {!r}".format(name))
    return "Power Switch" in name


def is_rpi_query(address, detect_logger):
    """Determines if address belongs to raspberry pi.

    Args:
        address (str): ip_address
        detect_logger (Logger): logs device interactions.
    Returns:
        bool: whether or not Raspberry Pi in response
    """
    try:
        name = host_utils.ssh_command(
            address,
            command=SSH_COMMANDS["RPI_PRODUCT_NAME"],
            user="pi",
            ssh_key_type="raspbian")
    except RuntimeError as err:
        detect_logger.info("is_rpi_query failure: " + repr(err))
        return False
    detect_logger.info("is_rpi_query response: {!r}".format(name))
    return "Raspberry Pi" in name


def is_unifi_query(address, detect_logger):
    """Determines if address belongs to unifi poe switch.

    Args:
        address (str): ip_address
        detect_logger (Logger): logs device interactions.
    Returns:
        bool: whether or not USW in response
    """
    try:
        product_name = host_utils.ssh_command(address,
                                              SSH_COMMANDS["UNIFI_PRODUCT_NAME"],
                                              user="admin",
                                              ssh_key_type="unifi_switch")
    except RuntimeError as err:
        detect_logger.info("is_unifi_query failure: " + repr(err))
        return False
    detect_logger.info("is_unifi_query response: {!r}".format(product_name))
    return "USW" in product_name


def usb_product_name_query(address, detect_logger):
    """Gets product name from usb_info.

    Args:
        address (str): communication_address
        detect_logger (Logger): logs device interactions.
    Returns:
        str: product_name or empty
    """
    product_name = usb_utils.get_product_name_from_path(address).lower()
    detect_logger.info("usb_product_name_query response: {}".format(product_name))
    return product_name


ADB_QUERY_DICT = {
}

DOCKER_QUERY_DICT = {
    DOCKER_QUERY.product_name: docker_product_name_query
}

GENERIC_QUERY_DICT = {
    GENERIC_QUERY.always_true: always_true_query
}

SERIAL_QUERY_DICT = {
    SERIAL_QUERY.product_name: usb_product_name_query
}

SSH_QUERY_DICT = {
    SSH_QUERY.is_dli: is_dli_query,
    SSH_QUERY.is_rpi: is_rpi_query,
    SSH_QUERY.is_unifi: is_unifi_query,

}


DETECT_DICTIONARY = {
    "AdbComms": ADB_QUERY_DICT,
    "DockerComms": DOCKER_QUERY_DICT,
    "SerialComms": SERIAL_QUERY_DICT,
    "SshComms": SSH_QUERY_DICT,
    "YepkitComms": GENERIC_QUERY_DICT
}


def determine_device_class(address, communication_type, log_file_path, create_switchboard_func):
    """Returns the device class(es) that matches the address' responses.

    Compares the device_classes DETECT_MATCH_CRITERIA to the device responses.

    Args:
        address (str): communication_address.
        communication_type (str): category of communication.
        log_file_path (str): local path to write log messages to.
        create_switchboard_func (func): function to create a switchboard.

    Returns:
        list: classes where the device responses match the detect criteria.
    """
    detect_logger = _setup_logger(log_file_path)
    try:
        device_classes = get_communication_type_classes(communication_type)
        return find_matching_device_class(
            address, communication_type, detect_logger, create_switchboard_func, device_classes)
    finally:
        file_handler = detect_logger.handlers[0]
        file_handler.close()
        detect_logger.removeHandler(file_handler)


def find_matching_device_class(address,
                               communication_type,
                               detect_logger,
                               create_switchboard_func,
                               device_classes):
    """Returns all classes where the device responses match the detect criteria.

    Args:
        address(str): communication_address.
        communication_type(str): category of communication.
        detect_logger(Logger): logs device interactions.
        create_switchboard_func (func): function to create a switchboard.
        device_classes(list): device_classes whose match criteria must be compared to.

    Returns:
        list: classes where the device responses match the detect criteria.
    """

    matching_classes = []
    responses = _get_detect_query_response(address,
                                           communication_type,
                                           detect_logger,
                                           create_switchboard_func)
    for device_class in device_classes:
        if _matches_criteria(responses, device_class.DETECT_MATCH_CRITERIA):
            matching_classes.append(device_class)
            detect_logger.info("{}: Match.".format(device_class.DEVICE_TYPE))
        else:
            detect_logger.info("{}: No Match.".format(device_class.DEVICE_TYPE))
    return matching_classes


def get_communication_type_classes(communication_type):
    """Returns classes with that communication type.

    Args:
        communication_type(str): category of communication.

    Returns:
        list: classes with that communication type.
    """
    all_classes = copy.copy(supported_classes.SUPPORTED_AUXILIARY_DEVICE_CLASSES)
    all_classes += copy.copy(supported_classes.SUPPORTED_PRIMARY_DEVICE_CLASSES)
    all_classes += copy.copy(supported_classes.SUPPORTED_VIRTUAL_DEVICE_CLASSES)
    matching_classes = []
    for device_class in all_classes:
        if device_class.COMMUNICATION_TYPE == communication_type:
            matching_classes.append(device_class)
    return matching_classes


def _get_detect_query_response(address,
                               communication_type,
                               detect_logger,
                               create_switchboard_func):  # pylint: disable=unused-argument
    """Gathers device responses for all queries of that communication type.

    Args:
        address (str): communication_address
        communication_type (str): category of communication.
        detect_logger (Logger): logs device interactions.
        create_switchboard_func (func): function to create switchboard with.

    Returns:
        dict: device responses keyed by query enum member.
    """
    query_dict = DETECT_DICTIONARY[communication_type]
    args = [address, detect_logger]

    query_responses = {}
    for query_name in query_dict:
        try:
            query_responses[query_name] = query_dict[query_name](*args)
        except Exception as err:
            detect_logger.debug(f"failed getting detect query response for {address}: {err!r}")
            query_responses[query_name] = repr(err)

    return query_responses


def _matches_criteria(responses, match_criteria):
    """Checks if response dict matches match criteria.

    Note:
        There are two categories of values in match_criteria: bool and regexp/str
        Bools must match exactly while regexp must find a match in the response value.

    Args:
        responses (dict): device responses keyed by query name.
        match_criteria (dict): match values keyed by query name

    Returns:
        bool: whether or not responses meets match criteria
    """
    for entry, value in match_criteria.items():
        if isinstance(value, bool):
            if responses[entry] != value:
                return False
        else:
            if not re.search(value, responses[entry]):
                return False
    return True


def _setup_logger(log_file_path):
    """Set up a logger to log device interactions to the detect file.

    Args:
        log_file_path (str): path to log to.

    Returns:
        logger: logger for logging to log_file_path.
    """
    detect_logger = logging.getLogger(log_file_path)
    detect_logger.setLevel(logging.DEBUG)
    handler = logging.FileHandler(log_file_path)
    formatter = logging.Formatter(FMT)
    handler.setFormatter(formatter)
    detect_logger.addHandler(handler)
    return detect_logger



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base class for all primary and virtual device classes."""
# pylint: disable=comparison-with-callable
import difflib
import functools
import inspect

import os
import re
import shutil
import time
import weakref

from gazoo_device import config
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.base_classes import first_party_device_base
from gazoo_device.switchboard import log_process
from gazoo_device.switchboard import parser
from gazoo_device.utility import build_utils
from gazoo_device.utility import supported_classes

logger = gdm_logger.get_gdm_logger()

# Number of times to retry urllib action (e.g. urlopen, urlretrieve...)
MAX_RETRY_URLLIB_ACTION = 3
ERROR_PREFIX = "Exception_"

TIMEOUTS = {
    "CONNECTED": 3,
    "SHELL": 60
}


def create_symlink(file_path, symlink_file_name):
    """Creates or updates a symlink symlink_file_name to the file_path specified.

    Args:
        file_path (path): to file to create symlink to
        symlink_file_name (str): to create or update in directory of file_path.

    Returns:
        str: Path to symlink file created or updated
    """

    directory = os.path.dirname(file_path)
    filename = os.path.basename(file_path)

    tmp_symlink_path = os.path.join(directory, symlink_file_name + ".tmp")
    symlink_path = os.path.join(directory, symlink_file_name)

    # We use a tmp name first in case this is already linked
    if os.path.lexists(tmp_symlink_path):
        os.remove(tmp_symlink_path)
    os.symlink(filename, tmp_symlink_path)
    os.rename(tmp_symlink_path, symlink_path)
    return symlink_path


def get_log_filename(log_directory, device_name, name_prefix=""):
    """Returns the full path of log filename using the information provided.

    Args:
        log_directory (path): to where the log file should be created.
        device_name (str): to use in the log filename
        name_prefix (str): string to prepend to the start of the log file.

    Returns:
        str: Path to log filename using the information provided.
    """

    log_timestamp = time.strftime("%Y%m%d-%H%M%S")
    if name_prefix:
        log_file_name = "{}-{}-{}.txt".format(name_prefix, device_name, log_timestamp)
    else:
        log_file_name = "{}-{}.txt".format(device_name, log_timestamp)
    return os.path.join(log_directory, log_file_name)


class GazooDeviceBase(first_party_device_base.FirstPartyDeviceBase):
    """Base class for all primary and virtual device classes."""
    _has_button_support = False
    # Paths to event filters relative to the filter directory
    _default_filters = []
    _CONNECTION_TIMEOUT = 3
    _COMMUNICATION_KWARGS = {}
    _OWNER_LDAP = ""  # override in child classes

    def __init__(self,
                 manager,
                 device_config,
                 log_file_name=None,
                 log_directory=None):
        self._log_object_lifecycle_event("__init__")
        self.manager_weakref = weakref.ref(manager)

        # Create a dictionary to store "properties".  For now keep the
        # classification of "persistent" and "optional".
        self.props = {'persistent_identifiers': device_config["persistent"],
                      'optional': device_config["options"]}

        self._make_device_ready_setting = device_config.get("make_device_ready", "on")
        if not isinstance(device_config.get("make_device_ready", "on"), str):
            raise errors.GazooDeviceError("Device creation failed. 'make_device_ready' "
                                          "should be a string. but instead it is a {}".
                                          format(str(type(device_config["make_device_ready"]))))
        self._make_device_ready_setting = self._make_device_ready_setting.lower()

        if self._make_device_ready_setting not in ["on", "off", "check_only"]:
            raise errors.GazooDeviceError("Device creation failed. 'make_device_ready' "
                                          "should be 'on', 'off' or 'check_only' not {}".
                                          format(device_config["make_device_ready"]))

        self._commands = {}
        self._regexes = {}
        self._timeouts = TIMEOUTS.copy()
        self.filter_paths = self._get_default_filters()
        self.filter_paths += device_config.get("filters") or []

        self._build_autocompleter = build_utils.BuildAutoCompleter(self.owner)
        if not hasattr(self, "_build_obtainer"):
            self._build_obtainer = build_utils.BuildObtainer(self.name, owner=self.owner)

        logger.debug("{} device_config: {}", self.name, self.props['persistent_identifiers'])
        logger.debug("{} device_options: {}", self.name, self.props['optional'])

        # Initialize log files
        self.log_directory = log_directory
        if log_file_name:
            self._log_file_name = os.path.join(log_directory, log_file_name)
        else:
            log_name_prefix = device_config["log_name_prefix"]
            self._log_file_name = get_log_filename(log_directory,
                                                   self.name,
                                                   name_prefix=log_name_prefix)
        self._update_event_filename_and_symlinks()
        if hasattr(self, "_COMMUNICATION_TYPE"):
            self.COMMUNICATION_TYPE = self._COMMUNICATION_TYPE
        if hasattr(self, "_DEVICE_TYPE"):
            self.DEVICE_TYPE = self._DEVICE_TYPE
        self.device_type = self.DEVICE_TYPE

    @decorators.SettableProperty
    def alias(self):
        """Returns the user-defined device alias (string)."""
        return self.props['optional']['alias']

    @decorators.PersistentProperty
    def commands(self):
        """Dictionary of commands issued to the device via shell."""
        return self._commands

    @decorators.PersistentProperty
    def communication_address(self):
        """Returns the address of the main communication port.

        Returns:
            str: path or address of main communication port.
        """
        name = self.props['persistent_identifiers'].get('console_port_name')
        if name:
            return name

        return self.props['persistent_identifiers'].get('adb_serial', 'None')

    @decorators.DynamicProperty
    def connected(self):
        """Returns whether or not device is connected."""
        device_config = {'persistent': self.props["persistent_identifiers"]}
        return self.is_connected(device_config)

    @decorators.PersistentProperty
    def health_checks(self):
        """Returns list of methods to execute as health checks."""
        return [self.check_usb_hub_ready,
                self.device_is_connected,
                self.check_create_switchboard]

    @decorators.DynamicProperty
    def log_file_name(self):
        """Returns current log file name in use.

        Returns:
            str: Path to current log file name.

        Note:
            When the device has been recently created it might be possible that
            the log file path does not yet exist but will be created very soon.
            The caller is still expected to check if the file path returned
            exists. The caller should refer to this property often because log
            rotation might cause the log path to change depending on the
            max_log_size value currently in use.
        """
        current_log_filename = self._log_file_name

        # Check if log file has rotated to next log filename
        next_log_filename = log_process.get_next_log_filename(current_log_filename)
        while os.path.exists(next_log_filename):
            current_log_filename = next_log_filename
            next_log_filename = log_process.get_next_log_filename(current_log_filename)
        return current_log_filename

    @decorators.PersistentProperty
    def model(self):
        return self.props['persistent_identifiers']['model']

    @decorators.PersistentProperty
    def name(self):
        return self.props['persistent_identifiers']['name']

    @decorators.PersistentProperty
    def owner(self):
        """LDAP associated with the owner of this device class."""
        return getattr(self, "_OWNER_LDAP", "")

    @decorators.PersistentProperty
    def serial_number(self):
        return self.props['persistent_identifiers'].get('serial_number')

    @property
    def event_parser(self):
        """Allows one to query events that have happened in the logs."""
        if not hasattr(self, "_event_parser"):
            self._event_parser = parser.Parser(
                self.filter_paths, device_name=self.name, event_file_path=self.event_file_name)
        return self._event_parser

    @decorators.PersistentProperty
    def regexes(self):
        """Regular expressions used to retrieve properties, events and states from device output.

        Returns:
            dict: mapping of name to regular expression.
        """
        return self._regexes

    @decorators.PersistentProperty
    def timeouts(self):
        """Dictionary of default timeouts to use when expecting certain actions."""
        return self._timeouts

    @decorators.PersistentProperty
    def build_properties(self):
        return self._build_autocompleter.get_default_values(self.DEVICE_TYPE)

    @decorators.LogDecorator(logger)
    def add_new_filter(self, filter_path):
        """Adds new log filter at path specified.

        Args:
            filter_path (str): filter file to add

        Raises:
            GazooDeviceError: if feature is not available for this device or
                             if filter_path doesn't exist
        """
        self.event_parser.load_filter_file(filter_path)
        self.switchboard.add_new_filter(filter_path)

    @decorators.health_check
    def check_create_switchboard(self):
        """Checks connection and creates switchboard through invocation."""
        self.switchboard.add_log_note("{} switchboard successfully started.".format(self.name))

    @decorators.health_check
    def device_is_connected(self):
        """Checks that device shows up as a connection on the host machine.

        Raises:
           DeviceNotConnectedError: if device is not connected.
        """
        logger.info(
            "{} waiting up to {}s for device to be connected.".format(
                self.name, self.timeouts["CONNECTED"]))
        end_time = time.time() + self.timeouts["CONNECTED"]
        while time.time() < end_time:
            if self.connected:  # pylint: disable=using-constant-test
                return
            time.sleep(.5)
        raise errors.DeviceNotConnectedError(
            self.name, msg="device not reachable from host machine.")

    @decorators.LogDecorator(logger, decorators.DEBUG)
    def close(self):
        """Calls close on the communication ports and resets anything needed.

        Note:
            Resets the buttons and terminates the child processes.
        """
        if hasattr(self, '_switchboard'):
            self.switchboard.close()
            self.switchboard.ensure_path_unlocked(self.name, self.communication_address)
            del self._switchboard
            self.reset_all_capabilities()

        if hasattr(self, '_event_parser'):
            del self._event_parser

        if hasattr(self, '_build_obtainer'):
            self._build_obtainer.close()
            del self._build_obtainer

        if hasattr(self, "manager_weakref"):
            manager_instance = self.manager_weakref()
            if manager_instance is not None:
                if self.name in manager_instance._open_devices:
                    self._log_object_lifecycle_event("close")
                    del manager_instance._open_devices[self.name]
            del manager_instance

    @decorators.LogDecorator(logger)
    def check_device_ready(self):
        """Checks if the device is healthy by executing a series of health check methods."""
        self._execute_health_check_methods(self.health_checks)

    @decorators.health_check
    def check_usb_hub_ready(self):
        """Check if the device's USB hub is configured and ready, if it exists."""
        if not self.is_detected():
            return
        if not hasattr(type(self), 'usb_hub'):
            return

        usb_hub_props = ["device_usb_hub_name", "device_usb_port"]
        unset_props = [prop for prop in usb_hub_props
                       if hasattr(type(self), prop) and not getattr(self, prop)]
        unset_persistent_props = [
            prop for prop in unset_props
            if isinstance(getattr(type(self), prop), decorators.PersistentProperty)]
        unset_settable_props = [
            prop for prop in unset_props
            if isinstance(getattr(type(self), prop), decorators.SettableProperty)]

        if unset_persistent_props or unset_settable_props:
            persistent_msg, settable_msg = "", ""
            if unset_persistent_props:
                persistent_msg = "use 'gdm redetect {}' to set {}".format(
                    self.name, ", ".join(unset_persistent_props))
            if unset_settable_props:
                settable_msg = "use 'gdm set-prop {} <property> <value>' to set {}".format(
                    self.name, ", ".join(unset_settable_props))
            if unset_persistent_props and unset_settable_props:
                how_to_fix_msg = f"{persistent_msg} and {settable_msg}"
            else:
                how_to_fix_msg = persistent_msg or settable_msg
            logger.info(f"{self.name} usb_hub capability is not available. "
                        f"If a USB hub is connected, {how_to_fix_msg}.")
            return

        try:
            logger.info("{} device usb port is set to {}".format(self.name,
                                                                 self.usb_hub.get_device_power()))
        except ValueError as err:
            logger.info("{} did not initiate usb_hub capability. Error: {!r}".format(self.name,
                                                                                     err))

    def get_dynamic_properties(self):
        """Returns a dictionary of prop, value for each dynamic property."""
        names = self.get_dynamic_property_names()
        return self._get_properties(names)

    def get_persistent_properties(self):
        """Returns a dictionary of prop, value for each persistent property."""
        names = self.get_persistent_property_names()
        return self._get_properties(names)

    def get_property(self, name, raise_error=False):
        """Retrieves a property value (can be nested).

        Args:
            name (str): name of a single property to fetch.
            raise_error (bool): raise error if unable to retrieve property

        Returns:
            object: value of the specified dynamic property.

        Raises:
            GazooDeviceError: property value is not a plain data object.
            AttributeError: property doesn't exist.
            Exception: property is not retrieveable for arbitrary reason.

        Note:
           Expects property value to be a plain data object.
           Returns a string when property doesn't exist or property raises an error.
           Able to process both device properties (firmware_version) and
           capability properties (wpan.ncp_address)
        """
        instance = self
        original_name = name
        try:
            if '.' in name:  # capability property
                instance = getattr(self, name.split('.')[0])
                name = name.split('.')[1]
            value = getattr(instance, name)
            if callable(value):
                raise errors.GazooDeviceError("{}'s {} is a method"
                                              .format(self.name, original_name))
            return value
        except AttributeError:
            if name in self.props['optional']:
                return self.props['optional'][name]
            if raise_error:
                raise
            close_matches = difflib.get_close_matches(name, self.get_property_names())
            return "{} does not have a known property '{}'. Close matches: {}".format(
                self.name,
                name,
                " or ".join(close_matches))
        except Exception as err:
            if raise_error:
                raise
            error_type = type(err).__name__
            logger.info("{} for {}, exception: {}".format(error_type, name, str(err)))
            return ERROR_PREFIX + error_type

    def get_property_names(self):
        """Returns a list of all property names."""
        full_list = list(self.get_dynamic_property_names())
        full_list.extend(self.get_persistent_property_names())
        full_list.extend(self.get_settable_property_names())
        full_list.extend(self.props['optional'])
        return list(set(full_list))

    def get_settable_properties(self):
        """Returns a dictionary of prop, value for each settable property."""
        names = self.get_settable_property_names()
        return self._get_properties(names)

    @decorators.LogDecorator(logger, decorators.DEBUG)
    def set_property(self, prop, value):
        """Set a settable property.

        Args:
          prop (str): property name
          value (object): value of the property name.

        Raises:
            ValueError: if property not settable.
        """
        if prop in self.get_persistent_property_names():
            raise ValueError("{}'s {} is a persistent property and not settable. "
                             "Redetect device if it's wrong.".format(self.name, prop))
        if prop in self.get_dynamic_property_names():
            raise ValueError("{}'s {} is a dynamic property and not settable. ".format(
                self.name, prop))
        self.props['optional'][prop] = value

    @decorators.LogDecorator(logger)
    def start_new_log(self, log_directory=None, log_name_prefix=""):
        """Start a new log and filter event file in the log directory specified.

        Args:
            log_directory (path): to where the new log file should be created.
            log_name_prefix (str): string to prepend to the start of the log file name.

        Note:
            If log_directory is not specified the new log file will be created
            in the same directory as the last log file.
        """
        if not log_directory:
            log_directory = self.log_directory
        new_log_filename = get_log_filename(log_directory,
                                            self.name,
                                            name_prefix=log_name_prefix)
        # Make sure new log file name is different from old log file name
        while self.log_file_name == new_log_filename:  # pylint: disable=comparison-with-callable
            time.sleep(0.5)
            new_log_filename = get_log_filename(log_directory,
                                                self.name,
                                                name_prefix=log_name_prefix)
        self.switchboard.start_new_log(new_log_filename)
        self.log_directory = log_directory
        self._log_file_name = new_log_filename
        if hasattr(self, "_event_parser"):
            del self._event_parser  # stops current log parser to allow update of event file.
        self._update_event_filename_and_symlinks()

    def get_firmware_version_type_or_skip(self, build_info):
        """Retrieves target firmware version and type from info and checks if version on device.

        Args:
            build_info (dict): dictionary of build information from parent method.

        Returns:
            tuple: firmware info as (str: target_version, str: target_firmware_type)

        Raises:
            GazooDeviceError: if unable to find file.
            SkipExceptionError: already on the target build.

        Notes:
            target_version may be:
                'UNKNOWN' if cannot regex version.
            target_firmware_type may be:
                'UNKNOWN' if build_info does not specify a type or if target_version is 'UNKNOWN'.
        """
        if 'build_file' in build_info:
            return "UNKNOWN", "UNKNOWN"  # unable to extract version from local file
        try:
            target_version = self._build_obtainer.get_version(self.DEVICE_TYPE, build_info)
        except Exception:
            return "UNKNOWN", "UNKNOWN"
        try:
            target_firmware_type = self._build_obtainer.get_firmware_type(
                self.DEVICE_TYPE, build_info)
        except Exception:
            target_firmware_type = "UNKNOWN"

        if not build_info.get('forced_upgrade'):
            logger.info("{} checking if build already on device".format(self.name))

            # Skip upgrade if current and target firmware is the same
            cur_firmware_version, cur_firmware_type = self._get_current_firmware()
            if self._is_on_firmware(cur_firmware_version,
                                    cur_firmware_type,
                                    target_version,
                                    target_firmware_type):
                raise decorators.SkipExceptionError("Already on {}{}".format(
                    target_firmware_type + " " if target_firmware_type != "UNKNOWN" else "",
                    target_version))

        return target_version, target_firmware_type

    def is_detected(self):
        """Returns whether or not persistent info has already been retrieved for device."""
        return bool(self.serial_number)

    @decorators.LogDecorator(logger, decorators.DEBUG)
    def make_device_ready(self, setting="on"):
        """Check device readiness and attempt recovery if allowed.

        Args:
            setting (str): 'on'|'off'|'check_only'|'flash_build'.

        Raises:
            GazooDeviceError: if device communication fails.

        Note:
            If setting is 'check_only', will skip recovery.
            If setting is 'off', will skip check_device_ready and recover.
            If setting is 'flash_build', will force upgrade if recovery is unsuccessful.
        """
        if setting == "off":
            return

        # check if the device is ready
        try:
            self.check_device_ready()
        except errors.GazooDeviceError as err:
            if setting == "check_only":
                logger.info("{} skipping device recovery", self.name)
                raise

            logger.info("{} failed check_device_ready with {}".format(self.name, repr(err)))

            # attempt to recover the device
            try:
                self.recover(err)
                logger.info("{} re-checking device readiness after recovery attempt", self.name)
                self.check_device_ready()
            except errors.GazooDeviceError as err:
                if setting != "flash_build":
                    raise

                # force upgrade the device
                logger.info("{} failed check_device_ready after recovery with {}".format(
                    self.name, repr(err)))
                logger.info("{} re-flashing device with a valid build".format(self.name))
                self.upgrade(forced_upgrade=True)

                # check if the device is ready after force upgrade
                logger.info("{} re-checking device readiness after flashing build", self.name)
                self.check_device_ready()
            logger.info("{} successfully recovered to ready state", self.name)

    def obtain_build(self, build_info=None):
        """Obtains build from known source and places it in /tmp/device_name.

        Args:
            build_info (dict): dictionary of build information from parent method.

        Raises:
            GazooDeviceError: if target file does not exist.

        Returns:
            str: local location of build file.

        Note:
            Automatically extracts .tgz files.
        """
        if not build_info:
            build_info = {}
        if build_info.get('method', 0) == 'dropbox':
            build_info['is_dropbox'] = True
            extract_build = False
        else:
            build_info['is_dropbox'] = False
            extract_build = build_info.get('extract_build', True)

        if 'build_file' in build_info:
            build_file = build_info['build_file']
            if not os.path.exists(build_file):
                raise errors.GazooDeviceError("Device {} upgrade failed. Unable to obtain build."
                                              " File {} does not exist".
                                              format(self.name, build_file))
            else:
                local_file_path = os.path.join(
                    self._build_obtainer.local_build_dir,
                    os.path.basename(build_file))
                if build_file != local_file_path:
                    shutil.copy(build_file, self._build_obtainer.local_build_dir)

        else:
            try:
                local_file_path = self._build_obtainer.locate_and_obtain_build(
                    self.DEVICE_TYPE, build_info)
            except Exception as err:
                raise errors.GazooDeviceError("Device {} upgrade failed. "
                                              "Unable to obtain build. Err: {!r}"
                                              .format(self.name, err))

        if extract_build:
            if local_file_path.endswith('.tgz'):
                local_file_path = self._build_obtainer.extract_tgz_build(local_file_path)
            elif local_file_path.endswith('.zip'):
                local_file_path = self._build_obtainer.extract_zip_build(local_file_path)
        return local_file_path

    @decorators.LogDecorator(logger)
    def verify_good_upgrade(self, target_firmware_version, target_firmware_type):
        """Verifies good upgrade by comparing expected firmware to current firmware.

        Args:
            target_firmware_version (str): firmware_version expected to be on device.
            target_firmware_type (str): firmware_type expected to be on device.

        Raises:
            GazooDeviceError: if target firmware does not match current firmware.

        Note:
            will skip target check if
                - target firmware version is unknown.
                - target firmware type is unknown or "unittest" and device supports firmware type.
        """
        supports_firmware_type = hasattr(self.__class__, "firmware_type")

        # determine if verification should be skipped
        if target_firmware_version == "UNKNOWN":  # unable to extract a version
            logger.info("{} target version {}. Skipping verification.".format(
                self.name, target_firmware_version))
            return
        elif target_firmware_type == "unittest" or (supports_firmware_type
                                                    and target_firmware_type == "UNKNOWN"):
            logger.info("{} target firmware type {}. Skipping verification.".format(
                self.name, target_firmware_type))
            return

        # verify current firmware matches target firmware
        cur_firmware_version, cur_firmware_type = self._get_current_firmware()
        if not self._is_on_firmware(cur_firmware_version,
                                    cur_firmware_type,
                                    target_firmware_version,
                                    target_firmware_type):
            raise errors.GazooDeviceError("Expected version {}{} but found {}{}".format(
                target_firmware_type + " " if supports_firmware_type else "",
                target_firmware_version,
                cur_firmware_type + " " if supports_firmware_type else "",
                cur_firmware_version))

    def _upgrade_wrapper(self, stage_number, transfer_and_recover_func, build_info):
        """Wraps around device-specific upgrade stages to obtain build and verify success.

        Args:
            stage_number (int): Total number of stages in upgrade. Varies between 2 and 5.
            transfer_and_recover_func (func): function to transfer build and recover from upgrade.
            build_info (dict): All the user given arguments

        Note:
            Before obtaining build, gets version and checks it against current device version.
            If version on device and not a forced_upgrade skips upgrade.
            If unable to get version from file, proceeds with upgrade and skips
                final version check.
            It leaves other stages like transferring the build, monitoring it, and recovering
                to the device class.
        """
        start_time = time.time()
        user_info = {name: value for name, value in build_info.items()
                     if value is not None and name != "self"}
        build_info_dict = {}
        default_values = self._build_autocompleter.get_default_values(self.device_type)
        build_info_dict.update(default_values)
        build_info_dict.update(user_info)  # user_info overwrites default values

        target_version, target_firmware_type = self.get_firmware_version_type_or_skip(
            build_info_dict)

        logger.info("{} Stage 1/{}: Obtain build", self.name, stage_number)

        local_file_path = self.obtain_build(build_info_dict)

        # Copy firmware image filename to device
        try:
            transfer_and_recover_func(local_file_path, build_info_dict)
        finally:
            try:
                if os.path.isdir(local_file_path):
                    shutil.rmtree(local_file_path)
                elif os.path.isfile(local_file_path):
                    os.remove(local_file_path)
            except Exception as err:
                logger.warning("{} Error while removing build file. Err: {!r}"
                               .format(self.name, err))

        # Verify firmware image is accepted by device and on new version
        self.verify_good_upgrade(target_version, target_firmware_type)
        upgrade_time = time.time() - start_time
        target_firmware_type_str = ""
        if target_firmware_type and target_firmware_type != "UNKNOWN":
            target_firmware_type_str = target_firmware_type + " "
        logger.info("{} successfully flashed with {}build {}", self.name, target_firmware_type_str,
                    target_version, upgrade_time)

    @classmethod
    def get_dynamic_property_names(cls):
        """Returns a list of dynamic property names including capability ones."""
        return cls._get_property_names(decorators.DynamicProperty)

    @classmethod
    def get_settable_property_names(cls):
        """Returns a list of settable property names including capability ones."""
        return cls._get_property_names(decorators.SettableProperty)

    @classmethod
    def get_persistent_property_names(cls):
        """Returns a list of persistent property names including capability ones."""
        names = cls._get_property_names(decorators.PersistentProperty)
        for property_value in config.CLASS_PROPERTY_TYPES:
            names += cls._get_property_names(property_value)
        return sorted(names)

    @classmethod
    def get_supported_capabilities(cls):
        """Returns a list of names of capabilities supported by this device class."""
        # Deduplicate names: there may be several flavors which share the same interface
        capability_names = {capability_class.get_capability_name()
                            for capability_class in cls.get_supported_capability_flavors()}
        return sorted(list(capability_names))

    @classmethod
    def get_supported_capability_flavors(cls):
        """Returns a set of all capability flavor classes supported by this device class."""
        capability_classes = [member.capability_classes for _, member in inspect.getmembers(cls)
                              if isinstance(member, decorators.CapabilityProperty)]
        return functools.reduce(set.union, capability_classes, set())

    @classmethod
    def has_capabilities(cls, capability_names):
        """Check whether this device class supports all of the given capabilities.

        Args:
            capability_names (list): names of capabilities to check for.

        Note:
            capability names are strings. They can be:
            - capability names ("keypad"),
            - capability interface names ("keypadbase"),
            - capability flavor names ("keypaddefault").
            If an interface name or capability name is specified, the behavior is identical:
                any capability flavor which implements the given interface will match.
            If a flavor name is specified, only that capability flavor will match.
            Different kinds of capability names can be used together (["wifi", "keypaddefault"]).

        Returns:
            bool: True if all of the given capabilities are supported by this device class,
                  False otherwise.

        Raises:
            GazooDeviceError: invalid type of capability_names argument OR
                             one of the capability names provided isn't recognized by GDM.
        """
        valid_capability_names_types = (list, tuple, set)
        if not isinstance(capability_names, valid_capability_names_types):
            raise errors.GazooDeviceError(
                "Invalid type of capability_names. Expected one of: {}, found: {}."
                .format(valid_capability_names_types, type(capability_names)))
        if not all(isinstance(cap_name, str) for cap_name in capability_names):
            raise errors.GazooDeviceError(
                "All capability names must be of string type. Found: {}.".format(capability_names))

        capabilities = []  # Interface or flavor classes
        for cap_name in capability_names:
            cap_name = cap_name.lower()
            if cap_name in supported_classes.SUPPORTED_CAPABILITY_FLAVORS:
                interface_or_flavor = supported_classes.SUPPORTED_CAPABILITY_FLAVORS[cap_name]
            elif cap_name in supported_classes.SUPPORTED_CAPABILITY_INTERFACES:
                interface_or_flavor = supported_classes.SUPPORTED_CAPABILITY_INTERFACES[cap_name]
            elif cap_name in supported_classes.SUPPORTED_CAPABILITIES:
                interface_name = supported_classes.SUPPORTED_CAPABILITIES[cap_name]
                interface_or_flavor = supported_classes.SUPPORTED_CAPABILITY_INTERFACES[
                    interface_name]
            else:
                msg = "\n".join([
                    "Capability {} is not recognized.".format(cap_name),
                    "Supported capability interfaces: {}".format(
                        supported_classes.SUPPORTED_CAPABILITY_INTERFACES.keys()),
                    "Supported capability flavors: {}".format(
                        supported_classes.SUPPORTED_CAPABILITY_FLAVORS.keys()),
                    "Supported capabilities: {}".format(
                        supported_classes.SUPPORTED_CAPABILITIES.keys())
                ])
                raise errors.GazooDeviceError(msg)
            capabilities.append(interface_or_flavor)

        supported_capabilities = cls.get_supported_capability_flavors()
        for requested_capability in capabilities:
            if not any(issubclass(supported_capability, requested_capability)
                       for supported_capability in supported_capabilities):
                return False
        return True

    def lazy_init(self, capability_class, *args, **kwargs):
        """Provides a lazy instantiation mechanism for capabilities.

        The capability instance will not be created until it is accessed for the first time.
        Subsequent accesses will return the same capability instance.
        In other words, no more than one capability instance will be created.
        If a capability is not accessed, the capability instance is not created.

        Args:
            capability_class (class): capability class to instantiate.
            *args (tuple): positional args to the capability's __init__.
                           Prefer using keyword arguments over positional arguments.
            **kwargs (dict): keyword arguments to the capability's __init__.

        Returns:
            CapabilityBase: initialized capability instance.
        """
        capability_name = self._get_private_capability_name(capability_class)
        if not hasattr(self, capability_name):
            capability_inst = capability_class(*args, **kwargs)
            setattr(self, capability_name, capability_inst)
        return getattr(self, capability_name)

    @decorators.LogDecorator(logger, decorators.DEBUG)
    def reset_all_capabilities(self):
        """Resets all capabilities which have been initialized by deleting them.

        Capabilities will be re-initialized on next use (when they're accessed).
        """
        for capability_class in self.get_supported_capability_flavors():
            self.reset_capability(capability_class)

    @decorators.LogDecorator(logger, decorators.DEBUG)
    def reset_capability(self, capability_class):
        """Resets the capability if it's already initialized by deleting it.

        The capability will be re-initialized on next capability use (when it's accessed).

        Args:
            capability_class (object): class of the capability to reset.
        """
        capability_name = self._get_private_capability_name(capability_class)
        if hasattr(self, capability_name):
            delattr(self, capability_name)

    def shell_with_regex(self,
                         command,
                         regex,
                         regex_group=1,
                         command_name="shell",
                         raise_error=False,
                         tries=2,
                         port=0,
                         searchwindowsize=config.SEARCHWINDOWSIZE,
                         check_return_code=False):
        """Sends a command, searches for a regex in the response, and returns a match group.

        Args:
            command (str): command to issue.
            regex (str): regular expression with one or more capturing groups.
            regex_group (int): number of regex group to return.
            command_name (str): command name to appear in log messages.
            raise_error (bool): whether or not to raise error if unable to find a match.
            tries (int): how many times to try executing the command before failing.
            port (int): which port to send the shell command to.
            searchwindowsize (int): Number of the last bytes to look at
            check_return_code (bool): whether to check the shell return code.

        Returns:
            str: value of the capturing group with index 'regex_group' in the match.

        Raises:
            GazooDeviceError: if command execution fails, shell return code is non-zero, or
                             couldn't find the requested group in any of the responses.
        """
        for _ in range(tries):
            try:
                response, code = self.shell(command,
                                            command_name=command_name,
                                            port=port,
                                            searchwindowsize=searchwindowsize,
                                            include_return_code=True)
            except errors.GazooDeviceError:
                continue

            if check_return_code and code != 0:
                logger.warning(
                    "{} shell responded with a non-zero return code. "
                    "Output: {}, return code: {}.".format(self.name, response, code))
                continue

            match = re.search(regex, response, re.MULTILINE | re.DOTALL)
            if match:
                max_group = match.lastindex
                if max_group is None:
                    max_group = 0

                if regex_group > max_group:
                    logger.warning(
                        "{}: requested group index ({}) exceeds index of last matched group ({}). "
                        "Matched groups: {}, response: {!r}, regex: {!r}."
                        .format(self.name, regex_group, max_group, match.groups(), response,
                                regex))
                else:
                    return str(match.group(regex_group))
            else:
                logger.debug("{} couldn't find {!r} in {!r}".format(self.name, regex, response))

        msg = "{} unable to retrieve {!r} from {!r} after {} tries".format(self.name, regex,
                                                                           command, tries)

        if raise_error:
            raise errors.GazooDeviceError(msg)
        else:
            logger.warning(msg)
            return ""

    @property
    def switchboard(self):
        """Instance for communicating with the device."""
        if not hasattr(self, "_switchboard"):
            switchboard_kwargs = self._COMMUNICATION_KWARGS.copy()
            switchboard_kwargs["communication_address"] = self.communication_address
            switchboard_kwargs["communication_type"] = self.COMMUNICATION_TYPE
            switchboard_kwargs["log_path"] = self.log_file_name
            switchboard_kwargs["device_name"] = self.name
            switchboard_kwargs["event_parser"] = self.event_parser
            self._switchboard = self.manager_weakref().create_switchboard(**switchboard_kwargs)

        return self._switchboard

    def _log_object_lifecycle_event(self, method_name):
        """Logs a message about a lifecycle event of a python object.

        Args:
            method_name (str): name of the method called on the object.
        """
        logger.debug("{} called on {} (id = {}) in process {}.".format(
            method_name, self, id(self), os.getpid()))

    def _execute_health_check_methods(self, health_checks):
        """Execute health checks on the device.

        Args:
            health_checks (list): list of methods to execute as health checks.

        Raises:
            CheckDeviceReadyError: if health check fails.

        Note:
            Order of health check methods matter. Health check execution will stop at the first
            failing health check.
        """
        checks_passed = []
        for step, health_check_method in enumerate(health_checks):
            method_name = health_check_method.__name__
            health_check_name = method_name.replace("_", " ").strip().capitalize()

            try:
                health_check_method()
            except errors.CheckDeviceReadyError as err:
                logger.info("{} health check {}/{} failed: {}.".format(
                    self.name, step + 1, len(health_checks), health_check_name))
                err.checks_passed = checks_passed
                err.properties = self.props["persistent_identifiers"].copy()
                raise

            checks_passed.append("{}.{}".format(type(self).__name__, method_name))
            logger.info("{} health check {}/{} succeeded: {}.".format(
                self.name, step + 1, len(health_checks), health_check_name))

    def _get_current_firmware(self):
        """Retrieve the current firmware version and type.

        Returns:
            tuple: (str: current firmware version, str: current firmware type)

        Raises:
            GazooDeviceError: if firmware properties could not be retrieved.
        """
        try:
            cur_firmware_version = self.firmware_version
        except Exception as err:
            raise errors.GazooDeviceError("{} unable to get firmware version. Error {!r}".format(
                self.name, err))

        try:
            # not all devices support firmware type
            # use getattr to skip device classes that don't have a firmware_type property
            cur_firmware_type = getattr(self, "firmware_type", None)
        except Exception as err:
            raise errors.GazooDeviceError("{} unable to get firmware type. Error {!r}".format(
                self.name, err))

        return cur_firmware_version, cur_firmware_type

    @classmethod
    def _get_default_filters(cls):
        """Returns the list of full paths to default device event filters.

        Returns:
            list: list of full paths (strings) to default device event filters.
        """
        filters = [os.path.join(config.FILTER_DIRECTORY, filter_path)
                   for filter_path in cls._default_filters]

        if not filters:
            logger.warning("Class {} has no default event filters defined.".format(cls.__name__))

        return filters

    @classmethod
    def _get_private_capability_name(cls, capability_class):
        """Returns the name of the private attribute to use for a given capability class.

        Args:
            capability_class (class): capability class object.

        Returns:
            str: name of the private capability attribute to be used.
        """
        return "_{}".format(capability_class.get_capability_name())

    def _get_properties(self, property_names):
        """Returns a dictionary of prop, value for each property."""
        property_dict = {}
        for name in property_names:
            value = self.get_property(name)
            if isinstance(value, str) and "does not have a known property" in value:
                continue  # property not supported in current flavor
            property_dict[name] = value

        return property_dict

    @classmethod
    def _get_property_names(cls, property_type):
        """Returns the property names for all public properties of the type."""
        classes = cls.get_supported_capability_flavors().copy()
        classes.add(cls)
        property_names = []
        for a_class in classes:
            prefix = "" if a_class == cls else a_class.get_capability_name() + "."
            for name, member in inspect.getmembers(a_class):
                if isinstance(member, property_type) and not name.startswith("_"):
                    property_names.append("{}{}".format(prefix, name))
        return list(set(property_names))

    def _is_on_firmware(self,
                        cur_firmware_version,
                        cur_firmware_type,
                        target_firmware_version,
                        target_firmware_type):
        """Determines if device is on expected firmware.

        Args:
            cur_firmware_version (str): current firmware version on the device.
            cur_firmware_type (str): current firmware type on the device, or None if device does
                                     not support firmware type.
            target_firmware_version (str): expected firmware version.
            target_firmware_type (str): expected firmware type.

        Returns:
            bool: if current firmware matches expected.
        """
        is_on_firmware_version = target_firmware_version in cur_firmware_version
        is_on_firmware_type = (not cur_firmware_type
                               or cur_firmware_type.lower() == target_firmware_type.lower())

        return is_on_firmware_version and is_on_firmware_type

    def _update_event_filename_and_symlinks(self):
        self.log_file_symlink = None
        if self.log_directory == config.DEFAULT_LOG_DIRECTORY:
            log_symlink_name = "{}-latest.txt".format(self.name)
            self.log_file_symlink = create_symlink(self._log_file_name, log_symlink_name)

        self.event_file_name = log_process.get_event_filename(self._log_file_name)
        self.event_file_symlink = None
        if self.log_directory == config.DEFAULT_LOG_DIRECTORY:
            event_symlink_name = "{}-latest-events.txt".format(self.name)
            self.event_file_symlink = create_symlink(self.event_file_name, event_symlink_name)

    def __del__(self):
        self._log_object_lifecycle_event("__del__")
        if hasattr(self, 'close'):
            self.close()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for log parser."""
from __future__ import absolute_import
from __future__ import division
import codecs
import os
import time

from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.switchboard import data_framer
from gazoo_device.switchboard import log_process


logger = gdm_logger.get_gdm_logger('log_parser')

DISPLAY_REFRESH = 3.0  # secs


class LogParser(object):
    """Provides ability to search for specific events in log file."""

    def __init__(self,
                 parser_obj,
                 log_path,
                 display_refresh=DISPLAY_REFRESH):
        """Initialize LogParser class using provided information.

        Args:
            parser_obj (Parser): Instance of class Parser
            log_path (str): Path to log filename containing raw, log event data
            display_refresh (float): Number of seconds to wait prior to refresh of display

        Raises:
            GazooParserError: If log_path does NOT exist
                             If event_filename already exists
                             If parser_object is None

        Note:
             Since the provided log_path is immediately parsed, initializing
        LogParser using log files exceeding 100 MB can cause test applications
        to appear to be delayed.  The smaller the log file the faster initialization
        will be.
        """

        if parser_obj is None:
            raise errors.GazooParserError("Log parser parameter check failed. "
                                          "Bad parser_obj.")

        if not os.path.isfile(log_path):
            raise errors.GazooParserError("LogParser parameter check failed. "
                                          "log file name: {} does not exist.".
                                          format(log_path))

        self.event_filename = log_process.get_event_filename(log_path)
        parser_obj.event_file_path = self.event_filename
        if os.path.isfile(self.event_filename):
            raise errors. GazooParserError("LogParser parameter check failed. "
                                           "event_filename: {} already exists.".
                                           format(self.event_filename))
        if display_refresh < 0:
            raise errors.GazooParserError("LogParser parameter check failed. "
                                          "Expected display refresh >=0 instead got: {}".
                                          format(display_refresh))
        self._parser_obj = parser_obj
        self._parse_events(log_path, display_refresh)

    def get_last_event(self, event_labels=None, timeout=1.0):
        r"""Returns the most recent matching event for each item in the list of event labels.

        Args:
            event_labels (list): list of event labels to lookup (example: ["power.state",
                "power.lost"]). None (default) returns the most recent event regardless of
                label.
            timeout (float): Timeout value in seconds. Example: 1.5.

        Raises:
            GazooParserError: on error parsing event file.

        Returns:
            ParserResult: An object containing the most recent event match for each event
            label in the event_labels list, or the most recent event if
            event_labels is None. The event data is stored in the .results_list
            attribute, and the count of results is in the .count attribute. If no matching
            events are found, the .results_list attribute will be set to the empty
            list.

            If the .timedout attribute is True, the operation timed out before all events
            were searched or the event file was not found before the timeout expired.

        Notes:
            The event_label is the desired filter filename without extension a . and the filter
            name (e.g. power.state).

            The following examples use a hypothetical power.json filter file with the
            following filters:

            .. code-block:: none

                {
                    "name": "state",
                    "regex_match": "power:(\\w+)"
                },
                {
                    "name": "lost",
                    "regex_match": "power lost"
                },
                {
                    "name": "restored",
                    "regex_match": "power restored"
                }

            The original log file being read looks like this:

            .. code-block:: none

                <2018-02-02 10:32:07.167234> [APPL] Spoke: power restored, powering on
                <2018-02-02 12:00:57.154328> [APPL] Spoke: power lost, powering down
                <2018-02-02 12:00:57.154328> power:OFF
                <2018-02-02 12:01:22.122824> [APPL] Spoke: power restored, powering on
                <2018-02-02 12:01:22.122824> power:ON
                <2018-02-02 12:02:18.121948> [APPL] Spoke: power lost, powering down
                <2018-02-02 12:02:18.121948> power:OFF

            Example output with event_labels=["power.state"], showing group matches:

            .. code-block:: none

                result.timedout = False
                result.results_list =
                    [{'system_timestamp': datetime.datetime(2018, 2, 2, 12, 2, 18, 121948),
                      'raw_log_line': 'power:OFF',
                      'power.state': ['OFF']}]
                result.count = 1

            Example output with event_labels=["power.lost"], showing matches with no groups:

            .. code-block:: none

                result.timedout = False
                result.results_list =
                    [{'system_timestamp': datetime.datetime(2018, 2, 2, 12, 2, 18, 121948),
                      'raw_log_line': '[APPL] Spoke: power lost',
                      'power.lost': []}]
                result.count = 1

            Example output with event_labels=["power.state", "power.lost", "power.restored"]:

            .. code-block:: none

                result.timedout = False
                result.results_list =
                    [{'system_timestamp': datetime(2018, 2, 2, 12, 2, 18, 121948),
                      'raw_log_line': 'power:OFF',
                      'power.state': ['OFF']},
                     {'system_timestamp': datetime(2018, 2, 2, 12, 2, 18, 121948),
                      'raw_log_line': '[APPL] Spoke: power lost',
                      'power.lost': []}
                     {'system_timestamp': datetime(2018, 2, 2, 12, 1, 22, 122824),
                      'raw_log_line': '[APPL] Spoke: power restored, powering on',
                      'power.restored': []}]
                results.count = 3
        """

        if event_labels:
            if not isinstance(event_labels, list):
                raise errors.GazooParserError("Expecting event_labels to be a list "
                                              "or None but received {} instead.".
                                              format(type(event_labels)))
            cur_event_labels = self.get_event_labels()
            for event_label in event_labels:
                if event_label not in cur_event_labels:
                    raise errors.GazooParserError("Event label {} doesn't exist.".
                                                  format(event_label))

        return self._parser_obj.get_last_event(self.event_filename,
                                               event_labels,
                                               timeout=timeout)

    def get_event_history(self, event_labels=None, count=None, timeout=10.0):
        r"""Returns up to count elements of event data matching given event labels.

        Args:
            event_labels (list): list of event labels to lookup (example: ["power.state",
                "power.lost"]). If None (default), return all events.
            count (int): of event data elements to return (Default: None for all events).
            timeout (float): Timeout value in seconds. Example: 1.5.

        Raises:
            GazooParserError: event_labels is not type list or
                             label in event_labels doesn't exist or
                             on error parsing event file

        Returns:
            ParserResult: An object containing events matching any of the event labels
            provided in event_labels, or all available events if event_labels in
            None. The event data is stored in the .results_list attribute and the count of
            results in the .count attribute. If no matching events are found, the
            .results_list attribute will be set to the empty list.

            If the .timedout attribute is True, the operation timed out before all events
            were searched, or the event file was not found before the timeout expired.

        Notes:
            Event info is listed in most recent to least recent order.

            The following examples use a hypothetical power.json filter file with the
            following filters:

            .. code-block:: none

                {
                    "name": "state",
                    "regex_match": "power:(\\w+)"
                },
                {
                    "name": "lost",
                    "regex_match": "power lost"
                },
                {
                    "name": "restored",
                    "regex_match": "power restored"
                }

            The original log file being read looks like this:

            .. code-block:: none

                <2018-02-02 10:32:07.167234> [APPL] Spoke: power restored, powering on
                <2018-02-02 12:00:57.154328> [APPL] Spoke: power lost, powering down
                <2018-02-02 12:00:57.154328> power:OFF
                <2018-02-02 12:01:22.122824> [APPL] Spoke: power restored, powering on
                <2018-02-02 12:01:22.122824> power:ON
                <2018-02-02 12:02:18.121948> [APPL] Spoke: power lost, powering down
                <2018-02-02 12:02:18.121948> power:OFF

            Example output with event_labels ["power.state"], showing group matches:

            .. code-block:: none

                result.timedout = False
                result.results_list =
                    [{'system_timestamp': datetime.datetime(2018, 2, 2, 12, 2, 18, 121948),
                      'raw_log_line': 'power:OFF',
                      'power.state': ['OFF']},
                     {'system_timestamp': datetime.datetime(2018, 2, 2, 12, 1, 22, 122824),
                      'raw_log_line': 'power:ON',
                      'power.state': ['ON']},
                     {'system_timestamp': datetime.datetime(2018, 2, 2, 12, 0, 57, 154328),
                      'raw_log_line': 'power:OFF',
                      'power.state': ['OFF']}]
                result.count = 3

            Example output with event_labels ["power.lost"], showing matches with no groups:

            .. code-block:: none

                result.timedout = False
                result.results_list =
                    [{'system_timestamp': datetime.datetime(2018, 2, 2, 12, 2, 18, 121948),
                      'raw_log_line: '[APPL] Spoke: power lost, powering down',
                      'power.lost': []},
                     {'system_timestamp': datetime.datetime(2018, 2, 2, 12, 0, 57, 154328),
                      'raw_log_line': '[APPL] Spoke: power lost, powering down',
                      'power.lost': []}]
                result.count = 2

            Example output with event_labels ["power.restored"], showing matches with no
            groups and count=1:

            .. code-block:: none

                result.timedout = False
                result.results_list =
                    [{'system_timestamp': datetime.datetime(2018, 2, 2, 12, 1, 22, 122824),
                      'raw_log_line': '[APPL] Spoke: power restored, powering on',
                      'power.restored': []}
                result.count = 1

            Example output with event_labels ["power.lost", "power.restored"]:

            .. code-block:: none

                result.timedout = False
                result.results_list =
                    [{'system_timestamp': datetime.datetime(2018, 2, 2, 12, 2, 18, 121948),
                      'raw_log_line: '[APPL] Spoke: power lost, powering down',
                      'power.lost': []},
                     {'system_timestamp': datetime.datetime(2018, 2, 2, 12, 1, 22, 122824),
                      'raw_log_line': '[APPL] Spoke: power restored, powering on',
                      'power.restored': []},
                     {'system_timestamp': datetime.datetime(2018, 2, 2, 12, 0, 57, 154328),
                      'raw_log_line': '[APPL] Spoke: power lost, powering down',
                      'power.lost': []}
                     {'system_timestamp': datetime.datetime(2018, 2, 2, 10, 32, 7, 167234),
                      'raw_log_line': '[APPL] Spoke: power restored, powering on',
                      'power.restored': []}]
                 result.count = 4
        """

        if event_labels:
            if not isinstance(event_labels, list):
                raise errors.GazooParserError("Expecting event_labels to be a list "
                                              "or None but received {} instead.".
                                              format(type(event_labels)))
            cur_event_labels = self.get_event_labels()
            for event_label in event_labels:
                if event_label not in cur_event_labels:
                    raise errors.GazooParserError("Event label {} doesn't exist.".
                                                  format(event_label))

        return self._parser_obj.get_event_history(event_labels,
                                                  count=count,
                                                  timeout=timeout)

    def get_event_history_count(self, event_label, timeout=10.0):
        """Returns count of event history matching event label specified.

        Args:
            event_label (str): event label to lookup (e.g. power.state)
            timeout (float): Timeout value in seconds. Example: 1.5.

        Raises:
            GazooParserError: on error parsing event file or
                             event_label doesn't exist

        Returns:
            ParserResult: An object containing the count of results in the .count
            attribute. The .results_list field is set to the empty list.

            If the .timedout attribute is True, the operation timed out before all events
            were searched, or the event file was not found before the timeout expired.
        """

        if event_label not in self.get_event_labels():
            raise errors.GazooParserError("Event label {} doesn't exist.".
                                          format(event_label))

        return self._parser_obj.get_event_history_count(self.event_filename,
                                                        event_label,
                                                        timeout=timeout)

    def get_event_label_dict(self, pattern=None):
        """Takes an event label pattern, and returns a dictionary of matching event labels.

        Args:
            pattern (str): An event label pattern regular expression

        Returns:
            dict: containing keys (i.e. event labels), and values (i.e regular expressions)

        Raises:
            GazooParserError: if pattern provided is an invalid regular expression

        Note:
            The pattern regular expression provided will be matched strictly against
            the beginning of each event_label.
            If no pattern is provided, the _filters_dict is returned.
        """

        return self._parser_obj.get_event_label_dict(pattern=pattern)

    def get_event_labels(self, pattern=None):
        """Takes an event label pattern, and returns a list of matching event labels.

        Args:
            pattern (str): An event label pattern regular expression

        Returns:
            list: of matching event label strings

        Raises:
            GazooParserError: if pattern provided is an invalid regular expression

        Note:
            The pattern regular expression provided will be matched strictly against
            the beginning of each event_label.
            If no pattern is provided, all filter event labels are returned.
        """

        return self._parser_obj.get_event_labels(pattern=pattern)

    def _parse_events(self, log_path, display_refresh):
        """Parses log file searching for events depositing results into an event file.

        Args:
            log_path (str): Path to log filename containing raw, log event data
            display_refresh (float): Number of seconds to wait prior to refresh of display

        Raises:
            GazooParserError: if log parser fails.

        Note:
            With large log files, this process may take a large amount of time.
        """

        logger.info("Parsing log file {} into event file {}, please wait",
                    log_path,
                    self.event_filename)
        start_time = time.time()
        with codecs.open(self.event_filename, "a", encoding="utf-8") as event_file:
            log_filename = os.path.basename(log_path)
            with codecs.open(log_path, "r", encoding="utf-8", errors="replace") as log_file:
                log_file.seek(0, os.SEEK_END)
                total_bytes = log_file.tell()
                log_file.seek(0, os.SEEK_SET)
                process_time = start_time
                buffered_unicode = u""
                framer = data_framer.NewlineFramer()
                new_header_length = (log_process.HOST_TIMESTAMP_LENGTH +
                                     log_process.LOG_LINE_HEADER_LENGTH)
                old_header_length = 29
                try:
                    while True:
                        log_data = log_file.read(size=4096)
                        if time.time() - process_time > display_refresh:
                            process_time = time.time()
                            bytes_processed = log_file.tell()
                            logger.info("{:.2%} complete - bytes processed: {} of {}",
                                        bytes_processed / total_bytes,
                                        bytes_processed,
                                        total_bytes)
                        if not log_data:
                            break

                        log_lines = buffered_unicode + log_data
                        buffered_len = len(buffered_unicode)
                        buffered_unicode = u""
                        for log_line in framer.get_lines(log_lines, begin=buffered_len):
                            if log_line[-1] == "\n":
                                if "> GDM-" in log_line:
                                    header_length = new_header_length
                                else:
                                    header_length = old_header_length
                                self._parser_obj.process_line(event_file,
                                                              log_line,
                                                              header_length=header_length,
                                                              log_filename=log_filename)
                            else:
                                buffered_unicode += log_line
                except IOError as err:
                    logger.debug("log_parser encountered error: {!r}".format(err))
                    raise errors.GazooParserError("Log file processing failed. "
                                                  "IOError: {!r}".format(err))
        logger.info("Parsing log file {} into event file {} finished in {}s",
                    log_path, self.event_filename, time.time() - start_time)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of the switch_power_unifi_switch capability."""
import re

from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger

from gazoo_device.capabilities.interfaces import switch_power_base

logger = gdm_logger.get_gdm_logger()

TELNET_COMMANDS = {
    "ENTER_TELNET": {
        "ENTER_CMD": ["telnet localhost", "enable"],
        "ENTER_REGEX": [r"\s+>", r"\s+#"],
        "EXIT_CMD": ["exit", "exit"],
        "EXIT_REGEX": [r"\s+>", "Connection closed by foreign host"]
    },
    "ENTER_TELNET_CONFIG": {
        "ENTER_CMD": ["telnet localhost", "enable", "config"],
        "ENTER_REGEX": [r"\s+>", r"\s+#", r"\(Config\)#"],
        "EXIT_CMD": ["exit", "exit", "exit"],
        "EXIT_REGEX": [r"\s+#", r"\s+>", "Connection closed by foreign host"]
    },
    "ENTER_TELNET_PORT_CONFIG": {
        "ENTER_CMD": ["telnet localhost", "enable", "config"],
        "ENTER_REGEX": [r"\s+>", r"\s+#", r"\(Config\)#"],
        "EXIT_CMD": ["exit", "exit", "exit", "exit"],
        "EXIT_REGEX": [r"\(Config\)#", r"\s+#", r"\s+>", "Connection closed by foreign host"]
    }
}

OFF = "off"
ON = "on"


class SwitchPowerUnifiSwitch(switch_power_base.SwitchPowerBase):
    """Definition of the switch_power_unifi_switch capability."""
    _REQUIRED_COMMANDS = ["GET_PORT_MODE", "ADJUST_PORTS_MODE", "GET_ALL_PORTS_MODE"]
    _REQUIRED_REGEXS = ["GET_PORT_MODE_REGEX", "GET_ALL_PORTS_MODE_REGEX"]

    def __init__(
            self,
            device_name,
            command_dict,
            regex_dict,
            total_ports,
            send_and_expect_fn,
            send_fn):
        """Create an instance of the poe switch power capability.

        Args:
            device_name (str): Name of the device this capability is attached to.
            command_dict (dict): A dictionary containing the command used for each method.
            regex_dict (dict): A dictionary containing the Regex used to validate device states
                               or command success.
            total_ports (int): Total number of device ports.
            send_and_expect_fn (func): Switchboard send_and_expect method.
            send_fn (func): Switchboard send method.
        """

        super().__init__(device_name=device_name)
        self._command_dict = command_dict
        self._regex_dict = regex_dict
        self._total_ports = total_ports
        self._send_and_expect_fn = send_and_expect_fn
        self._send_fn = send_fn

        self.validate_required_keys(self._REQUIRED_COMMANDS, self._command_dict, "command_dict")
        self.validate_required_keys(self._REQUIRED_REGEXS, self._regex_dict, "regex_dict")

    @decorators.PersistentProperty
    def supported_modes(self):
        """Get the Poe switch power modes supported by the unifi_switch."""
        return [OFF, ON]

    @decorators.CapabilityLogDecorator(logger)
    def power_on(self, port):
        """Powers on the specified port.

        Args:
            port (int): device port number
        """
        port = int(port)
        self._validate_port("power_on", port)
        telnet_commands = self._create_port_config_command_and_regex(port)
        logger.debug(f"{self._device_name} Powering on unifi_switch port {port}")
        self._telnet_port_config_send(
            telnet_commands=telnet_commands,
            func=self._send_fn,
            func_args=(self._command_dict["ADJUST_PORTS_MODE"].format(mode="auto"),))
        self._verify_mode("power_on", port, ON)

    @decorators.CapabilityLogDecorator(logger)
    def power_off(self, port):
        """Powers off the specified port.

        Args:
            port (int): device port number
        """
        port = int(port)
        self._validate_port("power_off", port)
        telnet_commands = self._create_port_config_command_and_regex(port)
        logger.debug(f"{self._device_name} Powering off unifi_switch port {port}")
        self._telnet_port_config_send(
            telnet_commands=telnet_commands,
            func=self._send_fn,
            func_args=(self._command_dict["ADJUST_PORTS_MODE"].format(mode="shutdown"),))
        self._verify_mode("power_off", port, OFF)

    @decorators.CapabilityLogDecorator(logger)
    def set_mode(self, mode, port):
        """Sets the given Ethernet port to the mode specified.

        Args:
            mode (str): Mode to set the specified unifi_switch port.
                        e.g. 'off', 'on'
            port (int): The port to set.

        Note:
            Not all devices support 'charge'

        Raises:
          GazooDeviceError: invalid port, or mode.
        """
        self._validate_mode(mode)
        if mode == OFF:
            self.power_off(port)
        else:
            self.power_on(port)

    @decorators.CapabilityLogDecorator(logger)
    def set_all_ports_mode(self, mode):
        """Set all device ports to the specified mode.

        Args:
            mode (str): Mode to set all unifi_switch ports to. Valid modes are "on" (auto) or
            "off (shutdown)".
        """
        self._validate_mode(mode)
        for poe_port in range(1, self._total_ports + 1):
            self.set_mode(mode, poe_port)

    def get_mode(self, port):
        """Get mode of the specified port.

        Args:
            port (int): device port number

        Returns:
            str: mode of specified port, either 'on'(auto), 'passive24V', or 'off'(shutdown).
        """
        port = int(port)
        self._validate_port("get_mode", port)
        response = self._telnet_config_send(
            func=self._poe_telnet_send_and_expect,
            func_args=(
                [self._command_dict["GET_PORT_MODE"].format(port=port)],
                [self._regex_dict["GET_PORT_MODE_REGEX"].format(port=port)]),
            func_kwargs={"regex_group": 1})

        if response == 'Auto':
            return ON
        elif response == 'Shutdown':
            return OFF
        else:
            return response

    def get_all_ports_mode(self):
        """Get mode of all the device ports.

        Returns:
            list: a list of the all the ports mode.
        """
        result = self._telnet_config_send(
            func=self._poe_telnet_send_and_expect,
            func_args=(
                [self._command_dict["GET_ALL_PORTS_MODE"]],
                [self._regex_dict["GET_ALL_PORTS_MODE_REGEX"]]),
            func_kwargs={"regex_group": 1})
        ports_status = []
        for poe_port in range(1, self._total_ports + 1):
            match = re.search(
                self._regex_dict["GET_PORT_MODE_REGEX"].format(port=poe_port), result)
            if match:
                if match.group(1) == 'Auto':
                    ports_status.append(ON)
                elif match.group(1) == 'Shutdown':
                    ports_status.append(OFF)
                else:
                    ports_status.append(match.group(1))
            else:
                ports_status.append("")
        return ports_status

    def _create_port_config_command_and_regex(self, poe_port):
        """Create the port config command and regex for the specified poe_port."""
        telnet_commands = TELNET_COMMANDS["ENTER_TELNET_PORT_CONFIG"].copy()
        telnet_commands["ENTER_CMD"].append(fr'interface 0/{poe_port}')
        telnet_commands["ENTER_REGEX"].append(fr'\(Interface 0/{poe_port}\)#')
        return telnet_commands

    def _telnet_port_config_send(self, telnet_commands, func, func_args=(), func_kwargs=None):
        """Send commands at the telnet_port_config menu level."""
        if func_kwargs is None:
            func_kwargs = {}
        self._poe_telnet_send_and_expect(
            telnet_commands["ENTER_CMD"], telnet_commands["ENTER_REGEX"])
        func_result = func(*func_args, **func_kwargs)
        self._poe_telnet_send_and_expect(
            telnet_commands["EXIT_CMD"], telnet_commands["EXIT_REGEX"])
        return func_result

    def _telnet_config_send(self, func, func_args=(), func_kwargs=None):
        """Send commands at the telnet_config menu level."""
        if func_kwargs is None:
            func_kwargs = {}
        self._poe_telnet_send_and_expect(
            TELNET_COMMANDS["ENTER_TELNET_CONFIG"]["ENTER_CMD"],
            TELNET_COMMANDS["ENTER_TELNET_CONFIG"]["ENTER_REGEX"])
        func_result = func(*func_args, **func_kwargs)
        self._poe_telnet_send_and_expect(
            TELNET_COMMANDS["ENTER_TELNET_CONFIG"]["EXIT_CMD"],
            TELNET_COMMANDS["ENTER_TELNET_CONFIG"]["EXIT_REGEX"])
        return func_result

    def _telnet_send(self, func, func_args=(), func_kwargs=None):
        """Send commands at the telnet menu level."""
        if func_kwargs is None:
            func_kwargs = {}
        self._poe_telnet_send_and_expect(
            TELNET_COMMANDS["ENTER_TELNET"]["ENTER_CMD"],
            TELNET_COMMANDS["ENTER_TELNET"]["ENTER_REGEX"])
        func_result = func(*func_args, **func_kwargs)
        self._poe_telnet_send_and_expect(
            TELNET_COMMANDS["ENTER_TELNET"]["EXIT_CMD"],
            TELNET_COMMANDS["ENTER_TELNET"]["EXIT_REGEX"])
        return func_result

    def _validate_port(self, method_name, port):
        """Ensure port is a valid port number.

        Args:
            method_name(str): name of the method where _validate_port is being called
            port (int): device port number

        Raises:
             GazooDeviceError: if input port is not valid
        """
        if port is None or port <= 0 or port > self._total_ports:
            raise errors.GazooDeviceError("Device {} {} failed. "
                                          "Port {} is invalid.".
                                          format(self._device_name, method_name, port))

    def _validate_mode(self, mode):
        """Verify mode given resides in the valid mode list.

        Args:
            mode (str): mode e.g. 'off', 'on'

        Raises:
            GazooDeviceError: if mode given is not found in the valid mode list.
        """
        if mode not in list(self.supported_modes):
            raise errors.GazooDeviceError("Mode {} not found in supported modes{}".
                                          format(mode,
                                                 self.supported_modes))

    def _verify_mode(self, method_name, port, mode):
        """Ensure port is set to specified mode.

        Args:
            method_name(str): name of the method where _validate_port is being called
            port (int): device port number
            mode (str): mode of the specified port. Valid modes are 'auto' or 'shutdown'.

        Raises:
             GazooDeviceError: if port is not in specified mode
        """
        active_mode = self.get_mode(port)
        if active_mode != mode:
            raise errors.GazooDeviceError(f"Device {self._device_name} {method_name} failed. "
                                          f"Mode is {active_mode} not {mode}.")

    def _poe_telnet_send_and_expect(
            self,
            commands,
            loglines,
            timeout=30.0,
            command_tries=1,
            raise_for_timeout=True,
            regex_group=0):
        """Helper method to send multiple send_and_expect commands during a telnet session.

        Args:
            commands (list): list of commands to send via send_and_expect.
            loglines (list): list of regex expressions to match with the corresponding command.
            timeout (float): timeout of send and expect.
            command_tries (int): number of times to retry the command.
            raise_for_timeout(bool): raise exception in event of timeout of send_and_expect.
            regex_group (int): regext match group to return.

        Returns:
            str: match of specified regex group.

        Raises:
             GazooDeviceError: if each command doesn't have a corresponding regex.
        """
        if len(loglines) != len(commands):
            raise errors.GazooDeviceError(f"The number of commands {len(commands)} does not equal "
                                          f"the number of loglines {len(loglines)}. "
                                          f"The number of commands and loglines must be equal.")
        for cmd_num in range(len(commands)):
            response = self._send_and_expect_fn(
                command=commands[cmd_num],
                pattern_list=[loglines[cmd_num]],
                timeout=timeout,
                command_tries=command_tries,
                raise_for_timeout=raise_for_timeout)

        return response.match.group(regex_group)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""ethernet_switch_api_base default implementation."""
from __future__ import absolute_import

from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import ethernet_switch_api_base
from gazoo_device.utility import common_utils

logger = gdm_logger.get_gdm_logger()

TIMEOUTS = {
    "ENABLE": 5,
    "DISABLE": 5,
}


class EthernetSwitchApiDefault(ethernet_switch_api_base.EthernetSwitchApiBase):
    """ethernet_switch_api_base default implementation."""

    def __init__(self,
                 device_name,
                 address, port,
                 disable_port_method, enable_port_method, get_port_status_method,
                 pre_enable_port_method=None, post_enable_port_method=None,
                 pre_disable_port_method=None, post_disable_port_method=None):
        """Creates an instance of the EthernetSwitchApiDefault capability.

        Args:
            device_name (str): name of the device with this capability. Used for log and error
                messages.

            address (str): Address of the ethernet switch

            port (str): Port number on ethernet switch where the device is connected

            disable_port_method (func): Function that takes port as input param and disables the
                port

            enable_port_method (func): Function that takes port as input param and enables
                the port

            get_port_status_method (func): Function that takes port as input param and returns
                a str variable refering to the port status

            pre_enable_port_method (func): Function that needs to be executed before enabling the
                port

            post_enable_port_method (func): Function that needs to be executed after enabling the
                port

            pre_disable_port_method (func): Function that needs to be executed before disabling the
                port

            post_disable_port_method (func): Function that needs to be executed after disabling the
                port
        """
        super(EthernetSwitchApiDefault, self).__init__(device_name=device_name)

        self._address = address
        self._port = port

        self.disable_port_method = disable_port_method
        self.pre_disable_port_method = pre_disable_port_method
        self.post_disable_port_method = post_disable_port_method

        self.enable_port_method = enable_port_method
        self.pre_enable_port_method = pre_enable_port_method
        self.post_enable_port_method = post_enable_port_method

        self.get_port_status_method = get_port_status_method

    @decorators.CapabilityLogDecorator(logger)
    def disable(self, no_wait=False):
        """Disables ethernet_switch port for device's ethernet connection.

        Args:
            no_wait (bool): If True, returns before verifying connection status.

        Raises:
          GazooDeviceError: if ethernet connection has not been disabled.
        """
        if not self.status:
            logger.debug("{} Ethernet connection is already disabled.".format(self._device_name))
            return

        if callable(self.pre_disable_port_method):
            self.pre_disable_port_method()

        logger.warn("{} Disabling the ethernet connection by turning off ethernet port# {!r} on "
                    "ethernet switch {!r}.".format(self._device_name, self.port, self.address))
        self.disable_port_method(self.port)

        if not no_wait:
            try:
                common_utils.retry(
                    func=getattr,
                    func_args=[self, "status"],
                    is_successful=common_utils.not_func,
                    timeout=TIMEOUTS["DISABLE"],
                    interval=1,
                    reraise=False
                )
            except errors.CommunicationTimeoutError:
                raise errors.GazooDeviceError(
                    "{} Failed to verify that ethernet connection is disabled".format(
                        self._device_name))

        if callable(self.post_disable_port_method):
            self.post_disable_port_method()

    @decorators.CapabilityLogDecorator(logger)
    def enable(self, no_wait=False):
        """Enables ethernet_switch port for devices ethernet connection.

        Args:
            no_wait (bool): If True, returns before verifying connection status.

        Raises:
            GazooDeviceError: if ethernet connection has not been enabled.
        """
        if self.status:
            logger.debug("{} Ethernet connection is already enabled.".format(self._device_name))
            return

        if callable(self.pre_enable_port_method):
            self.pre_enable_port_method()

        logger.warn("{} Enabling the ethernet connection by turning on ethernet port# {!r} on "
                    "ethernet switch {!r}.".format(self._device_name, self.port, self.address))
        self.enable_port_method(self.port)

        if not no_wait:
            try:
                common_utils.retry(
                    func=getattr,
                    func_args=[self, "status"],
                    is_successful=bool,
                    timeout=TIMEOUTS["ENABLE"],
                    interval=1,
                    reraise=False
                )
            except errors.CommunicationTimeoutError:
                raise errors.GazooDeviceError(
                    "{} Failed to verify that ethernet connection is enabled".format(
                        self._device_name))

        if callable(self.post_enable_port_method):
            self.post_enable_port_method()

    @property
    def address(self):
        """Returns the IP address of the ethernet switch where device is conneced.

        Returns:
            str: ethernet switch ip address
        """
        return self._address

    @property
    def port(self):
        """Returns the port number on the ethernet switch where device is conneced.

        Returns:
            str: ethernet switch ip address
        """
        return self._port

    @property
    def status(self):
        """Current status of the ethernet switch port.

        Returns:
            bool: True if ethernet ethernet switch port is enabled, otherwise False
        """
        port_status = self.get_port_status_method(self.port)
        return port_status.lower() in ["on", "1", "enable"]



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines an ssh transport interface by extending the process_transport.py.

Assumes logcat is available on the device.
"""
from __future__ import absolute_import
from gazoo_device.switchboard import process_transport
from gazoo_device.utility import host_utils


class SSHTransport(process_transport.ProcessTransport):
    """Perform transport communication using the ssh to some device."""

    def __init__(self,
                 comms_address,
                 args=host_utils.DEFAULT_SSH_OPTIONS,
                 ssh_key_type=None,
                 log_cmd="",
                 auto_reopen=True,
                 open_on_start=True,
                 username="root"):
        """Initialize the Ssh Transport object with the given properties.

        Args:
            comms_address (str): ip address to use when connecting to device
            args (str): extra args for ssh command.
            ssh_key_type (str): type of ssh key (types found in config)
            log_cmd (str): additional commands to pass to ssh such as logcat.
            auto_reopen (bool): flag indicating transport should be reopened if
                                unexpectedly closed.
            open_on_start (bool): flag indicating transport should be open on
                                  TransportProcess start.
            username (str): username to log into device.

        """
        self.comms_address = comms_address
        args = host_utils.generate_ssh_args(
            comms_address,
            log_cmd,
            username,
            options=args,
            ssh_key_type=ssh_key_type)
        super(SSHTransport, self).__init__(command="ssh",
                                           args=args,
                                           auto_reopen=auto_reopen,
                                           open_on_start=open_on_start)

    def _is_ready_to_open(self):
        return host_utils.is_pingable(self.comms_address)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Definition of the FastbootDefault capability."""
import time

from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import fastboot_base
from gazoo_device.utility import adb_utils
from gazoo_device.utility import common_utils

logger = gdm_logger.get_gdm_logger()


class FastbootDefault(fastboot_base.FastbootBase):
    """Fastboot capability."""
    _REQUIRED_TIMEOUTS = ["ONLINE"]

    def __init__(self,
                 device_name,
                 communication_address,
                 fastboot_serial,
                 timeouts_dict,
                 add_log_note_fn,
                 check_device_ready_fn,
                 root_device_fn):
        """Initialize the fastboot capability.

        Args:
            device_name (str): name of the device using the capability.
            communication_address (str): the device serial.
            fastboot_serial (str): the device fastboot serial.
            timeouts_dict (dict): dictionary of timeouts.
            add_log_note_fn (func): add output to log file.
            check_device_ready_fn (func): verify the device is ready to recieve GDM commands.
            root_device_fn (func): roots device if not rooted.
        """
        super(FastbootDefault, self).__init__(device_name=device_name)
        self._communication_address = communication_address
        self._fastboot_serial = fastboot_serial
        self._timeouts_dict = timeouts_dict
        self._add_log_note_fn = add_log_note_fn
        self._check_device_ready_fn = check_device_ready_fn
        self._root_device_fn = root_device_fn

        self.validate_required_keys(self._REQUIRED_TIMEOUTS, self._timeouts_dict, "timeouts_dict")

    @decorators.CapabilityLogDecorator(logger)
    def enter(self):
        """Enters fastboot (bootloader) mode for updating device firmware.

        Raises:
            GazooDeviceError: if unable to boot device into fastboot mode.
        """
        if self._is_in_fastboot_mode():
            return

        output = adb_utils.enter_fastboot(self._communication_address)
        for line in output.splitlines():
            self._add_log_note_fn(line + "\n")
        if not self._is_in_fastboot_mode(timeout=self._timeouts_dict["ONLINE"]):
            raise errors.GazooDeviceError("{} failed to enter fastboot mode in {}s."
                                          .format(self._device_name,
                                                  self._timeouts_dict["ONLINE"]))

    @decorators.CapabilityLogDecorator(logger)
    def exit(self):
        """Exits fastboot (bootloader) mode after updating device firmware.

        Raises:
            GazooDeviceError if unable to reboot device out of fastboot mode
        """
        if not adb_utils.is_fastboot_mode(self._fastboot_serial):
            return

        self._add_log_note_fn("GDM triggered reboot via 'fastboot reboot'")
        start_time = time.time()
        output = adb_utils.exit_fastboot(self._fastboot_serial)
        for line in output.splitlines():
            self._add_log_note_fn(line + "\n")
        logger.info("{} fastboot reboot took {}s.",
                    self._device_name, time.time() - start_time)
        self._check_device_ready_fn()
        self._root_device_fn()

    def _is_in_fastboot_mode(self, timeout=None):
        """Returns if device in fastboot mode. Waits up to timeout time.

        We try verify whether the device is in fastboot mode in the {timeout} duration, with retry
        interval 0.5 sec.
        If timeout is None (not provided), there will be no retry.

        Args:
            timeout (float): optional, maximum wait time in seconds. If timeout is None, we will
                             not retry.

        Returns:
            bool: whether device is in fastboot mode.
        """
        # No retry when timeout is None.
        # (Setting timeout to 0.1 is equivalent to "no retry" since interval is 0.5 sec.)
        timeout = timeout or 0.1

        try:
            common_utils.retry(adb_utils.is_fastboot_mode,
                               func_args=(self._fastboot_serial,),
                               is_successful=bool,
                               timeout=timeout,
                               interval=0.5)
            logger.info("{} is in fastboot mode.", self._device_name)
            return True
        except errors.CommunicationTimeoutError:
            return False



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Ubiquiti UniFi PoE Switch."""
import time

from gazoo_device import detect_criteria
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.base_classes import auxiliary_device
from gazoo_device.utility import host_utils
from gazoo_device.capabilities import shell_ssh
from gazoo_device.capabilities import switch_power_unifi_switch

logger = gdm_logger.get_gdm_logger()
CLASS_NAME = "UnifiPoeSwitch"
MARKER = "--- GDM Log Marker ---"
LOGGING_FILE_PATH = "/var/log/messages"
LOG_MARKER_LINE_POS_OR_EMPTY = ('grep -n -e "{marker}" {file_path} | tail -n 1 | cut -d '
                                '":" -f 1').format(marker=MARKER, file_path=LOGGING_FILE_PATH)
LOG_MARKER_LINE_POS = ("line_num=$({cmd}); echo ${{line_num:=1}}".
                       format(cmd=LOG_MARKER_LINE_POS_OR_EMPTY))

COMMANDS = {
    "MODEL": "mca-cli-op info",
    "FIRMWARE_VERSION": "mca-cli-op info",
    "SERIAL_NUMBER": "mca-cli-op info",
    "DEVICE_NAME": "mca-cli-op info",
    "ADJUST_PORTS_MODE": "poe opmode {mode}",
    "GET_PORT_MODE": "show poe port 0/{port}",
    "GET_ALL_PORTS_MODE": "show poe port all",
    "REBOOT": "reboot",
    "NUMBER_OF_PORTS": "mca-cli-op info",
    "LOGGING": "tail -F -n +$({cmd}) {file_path}".format(cmd=LOG_MARKER_LINE_POS,
                                                         file_path=LOGGING_FILE_PATH),
    "INJECT_LOG_MARKER": "sh -c 'echo \"{marker}\" >> {file_path}'".format(
        marker=MARKER, file_path=LOGGING_FILE_PATH),
    "TELNET": "echo 'exit' | telnet localhost",
    "MAC_ADDRESS": "show hardware"
}

REGEXES = {
    "MODEL_REGEX": r'Model:\s+(\S+)',
    "FIRMWARE_VERSION_REGEX": r'Version:\s+(\S+)',
    "SERIAL_NUMBER_REGEX": r'MAC Address:\s+(\S+)',
    "DEVICE_NAME_REGEX": r'Model:\s+([^-]+)',
    "GET_PORT_MODE_REGEX": r'0/{port}\s+(Auto|Passive24V|Shutdown)',
    "GET_ALL_PORTS_MODE_REGEX": r'(.*)\(Config\)#',
    "NUMBER_OF_PORTS_REGEX": r'Model:\s+\D+(\d+)',
    "MAC_ADDRESS_REGEX": r'Burned In MAC Address\.+ (\S+)'
}

TIMEOUTS = {
    "SHELL": 10,
    "SHUTDOWN": 60,
    "ONLINE": 120
}


class UnifiPoeSwitch(auxiliary_device.AuxiliaryDevice):
    """Device class for a Ubiquiti UniFi PoE Switch."""
    COMMUNICATION_TYPE = "SshComms"
    DETECT_MATCH_CRITERIA = {detect_criteria.SSH_QUERY.is_unifi: True}
    DEVICE_TYPE = "unifi_switch"
    _COMMUNICATION_KWARGS = {"log_cmd": COMMANDS["LOGGING"],
                             "ssh_key_type": "unifi_switch",
                             "username": "admin"}

    def __init__(self,
                 manager,
                 device_config,
                 log_file_name=None,
                 log_directory=None):
        super().__init__(manager,
                         device_config,
                         log_file_name=log_file_name,
                         log_directory=log_directory)
        self._commands.update(COMMANDS)
        self._regexes.update(REGEXES)
        self._timeouts.update(TIMEOUTS)

    @decorators.PersistentProperty
    def ip_address(self):
        """IP address."""
        return self.communication_address

    @decorators.DynamicProperty
    def firmware_version(self):
        """Version of UniFi PoE Switch.

        Returns:
            str: UniFi PoE Switch firmware version.
        """
        return self.shell_with_regex(command=self.commands["FIRMWARE_VERSION"],
                                     regex=self._regexes["FIRMWARE_VERSION_REGEX"])

    @decorators.PersistentProperty
    def health_checks(self):
        """Returns list of methods to execute as health checks."""
        return [self.device_is_connected,
                self.check_create_switchboard,
                self.check_telnet_connect]

    @decorators.DynamicProperty
    def total_ports(self):
        """Gets the number of ports for the attached unifi_switch.

        Returns:
            int: number of device ports.
        """
        return int(self.shell_with_regex(command=self.commands["NUMBER_OF_PORTS"],
                                         regex=self._regexes["NUMBER_OF_PORTS_REGEX"]))

    @decorators.CapabilityDecorator(switch_power_unifi_switch.SwitchPowerUnifiSwitch)
    def switch_power(self):
        """switch_power capability to send telnet commands to the device.

        Returns:
            SwitchPowerUnifiSwitch: switch_power_unifi_switch capability.
        """
        return self.lazy_init(
            switch_power_unifi_switch.SwitchPowerUnifiSwitch,
            device_name=self.name,
            command_dict=self.commands,
            regex_dict=self.regexes,
            total_ports=self.total_ports,
            send_and_expect_fn=self.switchboard.send_and_expect,
            send_fn=self.switchboard.send)

    @decorators.CapabilityDecorator(shell_ssh.ShellSSH)
    def shell_capability(self):
        return self.lazy_init(shell_ssh.ShellSSH, self.switchboard.send_and_expect, self.name,
                              timeout=self.timeouts["SHELL"], tries=2)

    def shell(self,
              command,
              command_name="shell",
              timeout=None,
              port=0,
              include_return_code=False):
        """Sends command and returns response and optionally return code.

        Args:
            command (str): Command to send to the device.
            command_name (str): Identifier for command.
            timeout (float): Time in seconds to wait for device to respond.
            port (int): Which port to send on, 0 or 1.
            include_return_code (bool): flag indicating return code should be returned.

        Raises:
            GazooDeviceError: if communication fails.

        Note:
            Can try multiple times as connection can sometimes fail.
            See shell_capability init args for setting the number of retry attempts.

        Returns:
            str: If include_return_code is False return the device response to the command.
            tuple: If include_return_code is True return the device response and return code.
        """
        timeout = timeout or self.timeouts["SHELL"]
        return self.shell_capability.shell(
            command, command_name=command_name, timeout=timeout, port=port,
            include_return_code=include_return_code)

    def shell_with_regex(self,
                         command,
                         regex,
                         regex_group=1,
                         command_name="shell",
                         raise_error=False,
                         tries=1,
                         port=0):
        """Sends a command, searches for a regex in the response, and returns a match group.

        Args:
            command (str): command to issue.
            regex (str): regular expression with one or more capturing groups.
            regex_group (int): number of regex group to return.
            command_name (str): command name to appear in log messages.
            raise_error (bool): whether or not to raise error if unable to find a match.
            tries (int): how many times to try executing the command before failing.
            port (int): which port to send the shell command to.

        Returns:
            str: value of the capturing group with index 'regex_group' in the match.

        Raises:
            GazooDeviceError: if command execution fails OR
                             couldn't find the requested group in any of the responses.
        """
        return self.command_with_regex(command,
                                       regex,
                                       self.shell,
                                       regex_group=regex_group,
                                       raise_error=raise_error,
                                       tries=tries,
                                       command_name=command_name,
                                       port=port)

    @decorators.health_check
    def check_telnet_connect(self):
        """Verify GDM is able to establish a telnet connection on the unifi_poe_switch CLI.

        Raises:
            GazooDeviceError: if telnet command returns a non zero exit status.
            DeviceNotResponsiveError: if no response to telnet command before the timeout.
        """
        try:
            response, return_code = self.shell(self.commands["TELNET"],
                                               include_return_code=True)

            if return_code != 0:
                raise errors.GazooDeviceError(f'{self.name} command {self.commands["TELNET"]}'
                                              f' returned {return_code} exit status.')

        except Exception as err:
            raise errors.DeviceNotResponsiveError(
                self.name, "failed to establish telnet connection", timeout=self.timeouts["SHELL"],
                details=str(err))

    @decorators.LogDecorator(logger)
    def get_detection_info(self):
        """Gets the persistent and optional attributes of the unifi_poe_switch.

        Returns:
          tuple: (dict, dict) dictionary of persistent attributes,
                  dictionary of optional attributes.

        Notes:
            persistent: model, serial_number
        """
        self.props['optional'] = {}
        self.props["persistent_identifiers"]["model"] = self.shell_with_regex(
            command=self.commands["MODEL"],
            regex=self._regexes["MODEL_REGEX"])
        self.props["persistent_identifiers"]["serial_number"] = self.shell_with_regex(
            command=self.commands["SERIAL_NUMBER"],
            regex=self._regexes["SERIAL_NUMBER_REGEX"]).replace(":", "")
        return self.props["persistent_identifiers"], self.props["optional"]

    @classmethod
    def is_connected(cls, device_config):
        """Checks whether or not the unifi_poe_switch is connected.

        Args:
          device_config (dict): contains "persistent" dict

        Returns:
            bool: whether the device responds to a ping.
        """

        return host_utils.is_pingable(device_config["persistent"]["console_port_name"])

    @decorators.LogDecorator(logger)
    def reboot(self, no_wait=False, method="shell"):
        """Reboots the device.

        Verifies device fully boots up afterwards.

        Args:
            no_wait (bool): Return before reboot completes. Default: False
            method (str): reboot technique to use.
        """
        self._inject_log_marker()
        self.switchboard.add_log_note("GDM triggered reboot")
        self.shell(command=self.commands["REBOOT"])
        if not no_wait:
            self._verify_reboot()

    @decorators.LogDecorator(logger)
    def recover(self, error):
        """Attempts to recover device based on the type of error specified."""
        if isinstance(error, errors.DeviceNotResponsiveError):
            self.reboot()
        else:
            raise error

    def _verify_reboot(self):
        """Verifies reboot actually occurred."""
        self._ensure_device_goes_offline()
        self._ensure_device_is_online()

    def _ensure_device_goes_offline(self, timeout=None):
        """Ensure device is no longer pingable over ssh.

        Args:
            timeout (float): Time in seconds to wait for device to respond.

        Raises:
            GazooDeviceError: Deviced failed to go offline before the timeout
        """
        timeout = timeout or self.timeouts["SHUTDOWN"]
        start_time = time.time()
        max_disconnect_time = start_time + timeout
        count = 0
        while time.time() < max_disconnect_time:
            if not host_utils.is_pingable(self.ip_address):
                count += 1  # Ensure device is really offline not just a blip
            else:
                count = 0
            if count == 2:
                logger.info("{} offline in {}s.".format(
                    self.name,
                    int(time.time() - start_time)))
                # close ssh transport as the ssh connection is disconnected.
                self.switchboard.close_all_transports()
                time.sleep(5)  # to ensure offline
                return
            time.sleep(.5)

        raise errors.GazooDeviceError(
            "Failed to go offline within {}s.".format(timeout))

    def _inject_log_marker(self):
        """Add GDM log marker to /var/log/syslog. This is to prevent reading stale logs.

        Note:
            Device logs are read starting with the last log marker (if present).
        """
        self.shell(self.commands["INJECT_LOG_MARKER"])

    def _ensure_device_is_online(self, timeout=None):
        """Ensure device is online and configs are fully loaded.

        Args:
            timeout (float): Time in seconds to wait for device to respond.

        Raises:
            GazooDeviceError: Device failed to come online before the timeout.
        """
        timeout = timeout or self.timeouts["ONLINE"]
        start_time = time.time()
        max_disconnect_time = start_time + timeout
        while time.time() < max_disconnect_time:
            if host_utils.is_pingable(self.ip_address):
                # There's a delay between the device being responsive to ping
                # and being able to open SSH connections
                time.sleep(10)
                self.switchboard.open_all_transports()
                break
            time.sleep(.5)
        else:
            raise errors.GazooDeviceError(
                "{} failed to become pingable in {}s.".format(self.name, timeout))

        output = "Device still offline"
        while time.time() < max_disconnect_time:
            try:
                output, return_code = self.shell(
                    self.commands["TELNET"], include_return_code=True)
                if return_code == 0:  # command executed
                    logger.info(
                        "{} online in {}s".format(self.name, int(time.time() - start_time)))
                    return
            except errors.GazooDeviceError:
                logger.debug("{} failed to respond to {!r}.".format(
                    self.name, self.commands["TELNET"]), exc_info=True)
            time.sleep(.5)
        raise errors.GazooDeviceError(
            "Failed to come online and respond to {!r} in {}s. Response: {}".format(
                self.commands["TELNET"], timeout, output))



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Gazoo Device Manager (gdm) is a single utility for control of a variety of smart devices.

It manages communication and setup of Gazoo devices to simplify daily tasks (software updating
a unit, getting logs) as well as serving as a device driver for automated (python) tests.
Areas of focus: ease of use (including ease of installation, setup, updating), robustness,
and scalability.

Examples of functionality:
    * Detecting and setting up devices when they are attached (create config files automatically).
    * Setting up shared permissions/file locations so all users can access test hardware (if they
      can log onto the pc).
    * Sharing a single device UART for logging and commands, to allow logs to be streamed "live"
      during a test while still sending commands/responses during a test.

Command Line Interface (CLI):
    There are several groups of commands.
        "Manager" commands operate on local config files on the pc.
            The command may talk to devices (detect), but they don't change the state of the
            device (i.e. read-only)
            The command may change the config files stored on the pc, to detect new devices,
            add an alias, etc.
        "Device" commands talk to or modify a Gazoo device.
            This includes upgrade, reboot, etc.
            For manager commands, the format is "gdm <command> <device> [<additional arguments>]".
            For example: gdm log cambrionix-1234.
            For device commands, the format is "gdm issue <device> - <command> [<arguments>]".
            For example: gdm issue cambrionix-1234 - reboot.

    You can get more details on a particular command at the command line with
    "gdm SomeCmd -h"
    For example::  gdm upgrade -h

Python API:
    ToDo: Add more API overview docs...
    Note: help(someCmd) will show the docstring for that command.
    Please comment about commands that have a confusing/missing
    docstring and we will improve it.

"""
from __future__ import absolute_import
import gc
import logging
import multiprocessing
import os
import signal
import sys
from gazoo_device import gdm_logger
from gazoo_device.base_classes import *  # pylint: disable=wildcard-import
from gazoo_device.primary_devices import *  # pylint: disable=wildcard-import
from gazoo_device.device_detector import DeviceDetector
from gazoo_device.manager import Manager
import gazoo_device.utility.common_utils
import gazoo_device.utility.reflection_utils

from gazoo_device._version import version
__version__ = version

# Defend against inadvertent basicConfig, which adds log noise
logging.getLogger().addHandler(logging.NullHandler())


# Ensure that atexit handlers run when killed by SIGTERM
def graceful_exit(*args, **kwargs):  # pylint: disable=unused-argument
    raise SystemExit(0)


signal.signal(signal.SIGTERM, graceful_exit)


if sys.platform == "darwin" and sys.version_info >= (3, 8):
    # Workaround for b/160958582: Python >= 3.8 defaults to "spawn" start method on Macs.
    # GDM isn't compatible with "spawn" yet, so force "fork" method on Macs until then.
    multiprocessing.set_start_method("fork", force=True)


# Set up logger
gdm_logger.initialize_logger()


def _after_fork():
    """Re-enable garbage collection in both parent & child process."""
    gc.enable()


def _before_fork():
    """Collect garbage, disable periodic GC until after fork, and flush logger messages.

    Note:
        This ensures logging thread is not logging anything during os.fork() call.
        If logging thread is logging during os.fork() call, children may be forked with an
        acquired stdout lock, which will cause a deadlock when the child process finishes.
        Child processes attempt to flush the stdout buffer before exiting. If a child process
        is forked with an acquired stdout buffer lock, it will hang indefinitely when trying to
        flush stdout because it cannot acquire the lock.
    """
    gc.disable()
    gc.collect()
    gdm_logger.flush_queue_messages()


# Periodic GC is active in child processes created via os.fork.
# Trigger a collection in the parent process prior to forking to prevent parent process's
# garbage from being copied over to the child and later being collected in each process.
# See b/150018610.
gazoo_device.utility.common_utils.register_at_fork(before=_before_fork,
                                                   after_in_parent=_after_fork,
                                                   after_in_child=_after_fork)

# Dynamically identify all device classes and capabilities supported by GDM
gazoo_device.utility.reflection_utils.identify_all_supported_classes()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""CONFIG FILE."""
from __future__ import absolute_import
import os

SEARCHWINDOWSIZE = 2000  # Used as default size of searchwindow used in switchboard functions
CUR_DIRECTORY = "/gazoo/gdm"
TEMP_DIRECTORY = os.path.join(os.environ["HOME"], "gdm")
if os.path.exists(CUR_DIRECTORY):
    GAZOO_DEVICE_DIRECTORY = CUR_DIRECTORY
else:
    GAZOO_DEVICE_DIRECTORY = TEMP_DIRECTORY
ADB_BIN_PATH_CONFIG = "adb_path"

PACKAGE_PATH = os.path.dirname(os.path.abspath(__file__))
BIN_PATH = os.path.join(PACKAGE_PATH, "bin")
DEFAULT_LOG_DIRECTORY = os.path.join(GAZOO_DEVICE_DIRECTORY, "log")
CONFIG_DIRECTORY = os.path.join(GAZOO_DEVICE_DIRECTORY, "conf")
FILTER_DIRECTORY = os.path.join(PACKAGE_PATH, "filters")
TTY_DIRECTORY = os.path.join(GAZOO_DEVICE_DIRECTORY, "tty")
DETOK_DIRECTORY = os.path.join(GAZOO_DEVICE_DIRECTORY, "detok")
BACKUP_PARENT_DIRECTORY = os.path.join(CONFIG_DIRECTORY, "backup")
BOTO_DIRECTORY = os.path.join(GAZOO_DEVICE_DIRECTORY, "botos")
KEYS_DIRECTORY = os.path.join(GAZOO_DEVICE_DIRECTORY, "keys")
BUILD_DEFAULTS_DIRECTORY = os.path.join(PACKAGE_PATH, "build_defaults")

DEFAULT_DEVICE_FILE = os.path.join(CONFIG_DIRECTORY, "devices.json")
DEFAULT_OPTIONS_FILE = os.path.join(CONFIG_DIRECTORY, "device_options.json")
DEFAULT_TESTBEDS_FILE = os.path.join(CONFIG_DIRECTORY, "testbeds.json")
DEFAULT_GDM_CONFIG_FILE = os.path.join(CONFIG_DIRECTORY, "gdm.json")

DEFAULT_LOG_FILE = os.path.join(DEFAULT_LOG_DIRECTORY, "gdm.txt")

DEFAULT_BUILD_INFO_FILE = os.path.join(BUILD_DEFAULTS_DIRECTORY, "info.json")
DEVICE_SCRIPTS_FOLDER = os.path.join(PACKAGE_PATH, "device_scripts")

REQUIRED_FOLDERS = [
    GAZOO_DEVICE_DIRECTORY,
    CONFIG_DIRECTORY,
    TTY_DIRECTORY,
    DEFAULT_LOG_DIRECTORY,
    DETOK_DIRECTORY,
    BACKUP_PARENT_DIRECTORY,
    BOTO_DIRECTORY,
    KEYS_DIRECTORY
]

KEY_FLAVOR_SSH = "ssh"
KEY_FLAVOR_API = "api"
# Keys used by GDM, such as private SSH keys. Format of entries:
# "key_name": {"remote_filename": "foo_key",
#              "flavor": KEY_FLAVOR_API or KEY_FLAVOR_SSH,
#              "local_path": os.path.join(KEYS_DIRECTORY, "foo_key")}
KEYS = {
    "raspbian":
        {"remote_filename": "raspberrypi3_ssh_key",
         "local_path": os.path.join(KEYS_DIRECTORY, "raspberrypi3_ssh_key"),
         "flavor": KEY_FLAVOR_SSH},
    "raspbian_public":
        {"remote_filename": "raspberrypi3_ssh_key.pub",
         "local_path": os.path.join(KEYS_DIRECTORY, "raspberrypi3_ssh_key.pub"),
         "flavor": KEY_FLAVOR_SSH},
    "unifi_switch":
        {"remote_filename": "unifi_switch_ssh_key",
         "local_path": os.path.join(KEYS_DIRECTORY, "unifi_switch_ssh_key"),
         "flavor": KEY_FLAVOR_SSH},
    "unifi_switch_public":
        {"remote_filename": "unifi_switch_ssh_key.pub",
         "local_path": os.path.join(KEYS_DIRECTORY, "unifi_switch_ssh_key.pub"),
         "flavor": KEY_FLAVOR_SSH}
}


REQUIRED_FILES = [
    DEFAULT_DEVICE_FILE,
    DEFAULT_OPTIONS_FILE,
    DEFAULT_TESTBEDS_FILE,
    DEFAULT_GDM_CONFIG_FILE,
    DEFAULT_LOG_FILE,
    DEFAULT_BUILD_INFO_FILE
]

REQUIRED_HUB_ENTRY_ATTRIBUTES = [
    "device_type", "model", "serial_number"
]

HUB_OPTION_ATTRIBUTES = [
    "alias"
]

REQUIRED_DEVICE_ENTRY_ATTRIBUTES = [
    "device_type", "model", "serial_number"
]

DEVICE_OPTION_ATTRIBUTES = [
    "alias", "power_switch", "power_port", "usb_hub", "location", "usb_port"
]

DEVICES_KEYS = ["devices", "other_devices"]
OPTIONS_KEYS = ["device_options", "other_device_options"]
TESTBED_KEYS = ["testbeds"]

DEFAULT_BOTO = os.path.join(os.environ["HOME"], ".boto")

CLASS_PROPERTY_TYPES = (str, int, dict, type(None))



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Digital Loggers Web Power Switch."""
import requests

from gazoo_device import decorators
from gazoo_device import gdm_logger
from gazoo_device.base_classes import auxiliary_device
from gazoo_device.utility import http_utils
from gazoo_device.utility import host_utils
from gazoo_device.switchboard import log_process
from gazoo_device.capabilities import switch_power_dli_powerswitch
from gazoo_device import detect_criteria

logger = gdm_logger.get_gdm_logger()

CLASS_NAME = "DliPowerSwitch"
AUTH = requests.auth.HTTPDigestAuth("admin", "1234")
HTTP_FRAMEWORK = "http://{ip}/restapi/"
GET_CONFIG = HTTP_FRAMEWORK + "config/="

HEADERS = {
    "GET_PROP": {"Accept": "application/json"},
    "SET_PROP": {"Accept": "application/json",
                 "X-CSRF": "x",
                 "X-HTTP-Method": "PUT"},
}

COMMANDS = {
    "FIRMWARE_VERSION": GET_CONFIG + "version/",
    "SERIAL_NUMBER": GET_CONFIG + "serial/",
    "DEVICE_MANUFACTURER": GET_CONFIG + "brand_company_name/",
    "DEVICE_NAME": GET_CONFIG + "brand_name/",
    "ADJUST_PORTS_MODE": HTTP_FRAMEWORK + "relay/outlets/{}/state/"
}

REGEXES = {
    "COMMAND_RESPONSE": ['[', '"', ']']
}


class DliPowerSwitch(auxiliary_device.AuxiliaryDevice):
    """Device class for a DLI Power Switch."""
    COMMUNICATION_TYPE = "SshComms"
    DETECT_MATCH_CRITERIA = {detect_criteria.SSH_QUERY.is_dli: True}
    DEVICE_TYPE = "powerswitch"
    _MODEL = "LPC9"
    _NUMBER_OF_PORTS = 8
    _VALID_RETURN_CODES = [200, 206, 207]

    def __init__(self,
                 manager,
                 device_config,
                 log_file_name=None,
                 log_directory=None):
        super().__init__(manager,
                         device_config,
                         log_file_name=log_file_name,
                         log_directory=log_directory)
        self._regexes.update(REGEXES)
        self._commands.update(COMMANDS)

    @decorators.PersistentProperty
    def ip_address(self):
        """Global IP address."""
        return self.communication_address

    @decorators.DynamicProperty
    def firmware_version(self):
        """Version of Power Switch.

        Returns:
            str: Power Switch firmware version.
        """
        return self.write_command("GET",
                                  self.commands["FIRMWARE_VERSION"].format(ip=self.ip_address),
                                  HEADERS["GET_PROP"])

    @decorators.CapabilityDecorator(switch_power_dli_powerswitch.SwitchPowerDliPowerswitch)
    def switch_power(self):
        """Returns a switch_power object to send commands.

        Returns:
            object: switch_power_dli_powerswitch
        """
        return self.lazy_init(
            switch_power_dli_powerswitch.SwitchPowerDliPowerswitch,
            http_fn=self.write_command,
            ip_address=self.ip_address,
            command_dict=self.commands,
            headers_dict=HEADERS,
            device_name=self.name,
            total_ports=self._NUMBER_OF_PORTS)

    @decorators.PersistentProperty
    def health_checks(self):
        """Returns list of methods to execute as health checks."""
        return [self.device_is_connected]

    @decorators.LogDecorator(logger)
    def get_detection_info(self):
        """Gets the persistent and optional attributes of a DLI Power Switch.

        Returns:
          tuple: (dict, dict) dictionary of persistent attributes,
                  dictionary of optional attributes.

        Notes:
            persistent: model, serial_number
        """
        self.props['optional'] = {}
        self.props["persistent_identifiers"]["model"] = self._MODEL
        self.props["persistent_identifiers"]["serial_number"] = self.write_command(
            "GET",
            self.commands["SERIAL_NUMBER"].format(ip=self.ip_address),
            headers=HEADERS["GET_PROP"])
        return self.props["persistent_identifiers"], self.props["optional"]

    @classmethod
    def is_connected(cls, device_config):
        """Checks whether or not the DLI Power Switch is connected.

        Args:
          device_config (dict): contains "persistent" dict

        Returns:
            bool: whether the device responds to a ping.
        """

        return host_utils.is_pingable(device_config["persistent"]["console_port_name"])

    @decorators.LogDecorator(logger)
    def reboot(self):
        """Issues a soft reboot command.

        Raises:
            NotImplementedError: not supported.
        """
        raise NotImplementedError("{}: reboot is not supported by {}."
                                  .format(self.name, self.DEVICE_TYPE))

    @decorators.LogDecorator(logger)
    def recover(self, error):
        """Attempts to recover device based on the type of error specified."""
        pass

    def write_command(self, method, url, headers, data=None, json_data=None):
        """Sends an HTTP request via specified method.

        Args:
            method (str): Method of HTTP request. Choices: "GET", "POST".
            url (str): HTTP formated command to send to the device.
            headers (dict): Headers required for the HTTP request.
            data (dict): Data that is needed for the HTTP POST Request
            json_data (dict): JSON data that is needed for the HTTP POST Request

        Raises:
            RuntimeError: if response.status_code returned by requests.get or requests.post is not
                          in valid_return_codes
            TypeError: if headers / json_data is not a dictionary or None.

        Returns:
            str: Formatted GET/POST HTTP response

        """
        log_message = "Sending command '{command}' to {device} via http {method} method.".format(
            command=url, device=self.name, method=method)
        self._log_to_file(log_message)
        try:
            if "GET" in method:
                response = http_utils.send_http_get(
                    url,
                    auth=AUTH,
                    headers=headers,
                    valid_return_codes=self._VALID_RETURN_CODES)
            else:
                response = http_utils.send_http_post(
                    url,
                    auth=AUTH,
                    headers=headers,
                    data=data,
                    json_data=json_data,
                    valid_return_codes=self._VALID_RETURN_CODES)

        except (RuntimeError, TypeError) as err:
            log_message = "Command '{command}' failed".format(command=url)
            self._log_to_file(log_message)
            raise err

        log_message = "Command '{command}' sent successfully. Return Code: {return_code}".format(
            command=url,
            return_code=response.status_code)
        self._log_to_file(log_message)
        return self._format_response(response)

    @classmethod
    def _format_response(cls, response):
        """Removes extraneous formatting from requests library response.text.

        Args:
            response (requests.Response): requests library response object
        Returns:
            str: returns a formatted string of response.text
        """
        return "".join(i for i in response.text if i not in REGEXES["COMMAND_RESPONSE"])

    def _log_to_file(self, log_message, port="M"):
        """Adds a log message for each command."""
        if not log_message.endswith('\n'):
            log_message += '\n'
        with open(self.log_file_name, 'a') as out_file:
            logline = log_process._add_log_header(log_message, port=port)
            out_file.write(logline)

    def _list_properties_dynamic_powerswitch(self):
        """Returns a set of dynamic properties.

        Returns:
            set: dynamic properties that can be queried during get-prop CLI/API operation.
        """
        return set(["firmware_version"])



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Definition of SideloadDefault capability."""

from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import sideload_base
from gazoo_device.utility import adb_utils
from gazoo_device.utility import common_utils

logger = gdm_logger.get_gdm_logger()


class SideloadDefault(sideload_base.SideloadBase):
    """Default implementation for sideload capability."""

    WAIT_TIME_IN_SECS_FOR_ENTERING_SIDELOAD_MODE = 30.0

    def __init__(self,
                 device_name,
                 device_serial,
                 reboot_fn,
                 check_device_ready_fn,
                 root_device_fn):
        """Initialize the sideload capability.

        Args:
            device_name (str): the device name.
            device_serial (str): the device serial.
            reboot_fn (func): the reboot function to exit sideload.
            check_device_ready_fn (func): verify the device is ready to recieve GDM commands.
            root_device_fn (func): roots device if not rooted.
        """
        super(SideloadDefault, self).__init__(device_name=device_name)
        self._device_serial = device_serial
        self._reboot_fn = reboot_fn
        self._check_device_ready_fn = check_device_ready_fn
        self._root_device_fn = root_device_fn

    @decorators.CapabilityLogDecorator(logger)
    def enter(self, auto_reboot=False):
        """Enters sideload mode for updating device firmware.

        Args:
            auto_reboot (bool): whether to use sideload-auto-reboot or not.

        Raises:
            GazooDeviceError: if unable to boot device into sideload mode.
        """
        if self.is_in_sideload_mode():
            return
        _ = adb_utils.enter_sideload(adb_serial=self._device_serial, auto_reboot=auto_reboot)
        self._verify_in_sideload_mode()

    @decorators.CapabilityLogDecorator(logger)
    def exit(self):
        """Exits sideload mode after updating device firmware.

        Raises:
            GazooDeviceError if unable to reboot device out of sideload mode
        """
        if not self.is_in_sideload_mode():
            return
        self._reboot_fn()
        self._check_device_ready_fn()
        self._root_device_fn()

    @decorators.CapabilityLogDecorator(logger)
    def is_in_sideload_mode(self):
        """Whether the device is currently in sideload mode.

        Returns:
            bool: True if device in sideload mode; False otherwise.
        """
        return adb_utils.is_sideload_mode(adb_serial=self._device_serial)

    def _verify_in_sideload_mode(self, timeout=None):
        """Wait up to timeout for device to enter sideload mode.

        Args:
            timeout (float): timeout in seconds to wait for entering sideload.

        Raises:
            GazooDeviceError: if unable to verify in sideload mode until timeout is reached.
        """
        if timeout is None:
            timeout = self.WAIT_TIME_IN_SECS_FOR_ENTERING_SIDELOAD_MODE
        try:
            common_utils.retry(self.is_in_sideload_mode,
                               is_successful=bool,
                               timeout=timeout,
                               interval=0.5)
        except errors.CommunicationTimeoutError as err:
            raise errors.GazooDeviceError('Failed to verify sideload mode. {}'.format(err))



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""TcpTransport is used for connection-oriented transmissions.

TcpTransport is responsible for delivering data to the appropriate application process
on the host computer. A host and port number are combined to create a network socket (an
identification address of a process-to-process communication). The socket can
then used by an application to handle a data stream for various needs.

TcpTransport implements the standard GDM transport interface (TransportBase).
"""

from __future__ import absolute_import
import socket
import six
from gazoo_device.switchboard import transport_properties
from gazoo_device.switchboard import transport_base

CONNECT_TIMEOUT = 10.0  # in secs


def check_port(port):
    """Verifies port value given is valid.

    Args:
        port (int): port number to verify

    Raises:
        ValueError: if port number provided is invalid

    Returns:
        int: port number
    """

    if port < 0 or port > 65535:
        raise ValueError("Port {} out of range".format(port))
    return port


def check_host(host):
    """Verifies host given is valid.

    Args:
        host (str): Either IP Address or hostname to verify

    Raises:
        ValueError: if ip address provided is invalid

    Returns:
        str: host
    """

    try:  # Valid ip address?
        socket.inet_pton(socket.AF_INET, host)
    except socket.error:
        try:  # Valid hostname?
            socket.getaddrinfo(host, None)
        except socket.error:
            raise ValueError("Host Address {} not valid".format(host))
    return host


class TcpTransport(transport_base.TransportBase):
    """Transport to communicate to ISA3 boxes."""

    def __init__(self,
                 comms_address,
                 args,
                 connect_timeout=CONNECT_TIMEOUT,
                 auto_reopen=False,
                 open_on_start=False):
        """Initialize the TcpTransport object with the given TCP properties.

        Args:
            comms_address (str): hostname or IP address of the device to connect to.
            args (str): The port to communicate to.
            connect_timeout (float): timeout on the socket instance before attempting connect.
            auto_reopen (bool): flag indicating transport should be reopened if
                                unexpectedly closed.
            open_on_start (bool): flag indicating transport should be open on
                                  TransportProcess start.

        Raises:
            ValueError: host not valid or port out of range
        """
        self.comms_address = comms_address
        port = int(args)
        super(TcpTransport, self).__init__(auto_reopen, open_on_start)
        self._properties.update({
            transport_properties.HOST: comms_address,
            transport_properties.PORT: int(args),
            transport_properties.CONNECT_TIMEOUT: connect_timeout,
        })
        check_host(comms_address)
        check_port(port)
        self._socket = None

    def is_open(self):
        """Returns True if tcp socket has been created.

        Returns:
            bool: True if _socket is not None
        """
        return hasattr(self, "_socket") and self._socket is not None

    def _open(self):
        """Opens or reopens the tcp connection using the current property values.

        Raises:
            IOError: if host and port do not resolve
        """
        self._socket = socket.create_connection(
            (self._properties[transport_properties.HOST],
             self._properties[transport_properties.PORT]),
            self._properties[transport_properties.CONNECT_TIMEOUT])

    def _close(self):
        """Closes the TCP connection."""
        self._socket.shutdown(socket.SHUT_RDWR)
        self._socket.close()
        self._socket = None

    def _read(self, size=1, timeout=None):
        """Returns bytes read up to max_bytes within timeout in seconds specified.

        Args:
            size (int): maximum number of bytes to read within timeout seconds
            timeout (float): maximum seconds to wait to read bytes or indefinitely
                             if timeout=None. Default: None

        Returns:
            str: bytes read from transport
                 or "" if no bytes were read
                 or "" if timeout was reached
        """
        try:
            self._socket.settimeout(timeout)
            read_bytes = self._socket.recv(size)
            return "" if read_bytes is None else read_bytes
        except socket.timeout:
            return ""

    def _write(self, data, timeout=None):
        """Writes the command provided to the TCP socket.

        Args:
            data (str): bytes to be written out within timeout seconds
            timeout (float): maximum seconds to wait to write bytes or indefinitely
                             if timeout=None. Default: None

        Returns:
            int: number of bytes written or None if no bytes were written
        """
        if isinstance(data, six.text_type):
            data = data.encode("utf-8", errors="replace")

        self._socket.settimeout(timeout)
        self._socket.sendall(data)
        return len(data)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Gets information on all usb connected devices.

Returns a dictionary mapping a persistent address (adb or serial port) to usb info
for that address.

Note: usb_info_linux primarily relies on pyudev for full information.
It processes just adb devices and those that show up in /dev/serial/by-id
"""
import re
import sys
from gazoo_device.utility import usb_config
import pyudev

PERSISTENT_SERIAL_PATH_FOLDER = '/dev/serial/by-id'
MOUNTED_DISK_FOLDER = '/dev/disk/by-id'
ADB_SERIAL_REGEX = r"\/dev\/disk\/by-id\/usb-Linux_File-CD_Gadget_([\da-z]+)-0:0"
IDENTIFIER_REGEX = r"\d-\d(?:\.\d?)+(\d\.\d):\d\.\d"
ANDROID_DEVICE_PRODUCT_NAME = "File-CD Gadget"

# These vendor's mounted drives show up as multiple usb connections.
# We use the one with the symlink.
REPEAT_VENDORS = ["FTDI", "cambrionix", "Linux", "Nest", "Android"]


def get_address_to_usb_info_dict():
    """Returns a dict of address to the usb information.

    Returns:
       dict: dictionary of address to UsbInfo instance for each usb device attached.
    """
    udev_devices = get_pyudev_list_of_devices()
    relevant_devices = [device for device in udev_devices
                        if device.get('ID_BUS') == 'usb']
    address_to_usb_info_dict = {}
    location_dict = {}
    model_dict = {}
    for device in relevant_devices:
        address, entry, dev_path, model = _process_udev_device(device)
        if address:  # remove devices that don't have good addresses
            address_to_usb_info_dict[address] = entry
            if dev_path:
                location_dict[address] = dev_path
            if model:
                model_dict[address] = model
    disk_info = _get_disk_info()
    _add_disk_info(address_to_usb_info_dict, disk_info)
    _add_cambrionix_ports(address_to_usb_info_dict, location_dict, model_dict)
    return address_to_usb_info_dict


def get_pyudev_list_of_devices(subsystem=None, devtype=None):
    """Wrapper around pyudev.list_devices.

    Args:
        subsystem (str): maps to pyudev.Context().list_devices arg.
        devtype (str): maps to pyudev.Context().list_devices arg DEVTYPE.

    Returns:
        list: a list of pyudev devices.

    Raises:
        RuntimeError: if platform is not linux.
    """
    args = {}
    if subsystem:
        args['subsystem'] = subsystem
    if devtype:
        args['DEVTYPE'] = devtype
    if sys.platform == 'darwin':
        raise RuntimeError('This method is only supported on linux.')
    context = pyudev.Context()
    return list(context.list_devices(**args))


def _add_cambrionix_ports(address_to_usb_info_dict, location_dict, model_dict):
    """Add cambrionix port number and parent address to each cambrionix child."""
    for cam_address, model in model_dict.items():
        port_mapping = usb_config.CAMBRIONIX_PORT_MAP[model]
        for child_address in address_to_usb_info_dict[cam_address].child_addresses:
            if child_address in address_to_usb_info_dict:
                child_entry = address_to_usb_info_dict[child_address]
                child_entry.usb_hub_address = cam_address
                child_entry.usb_hub_port = _get_cambrionix_port_number(
                    location_dict[child_address],
                    port_mapping)


def _add_disk_info(address_to_usb_info_dict, disk_info):
    """Add disk path to the entries that have them."""
    for entry in address_to_usb_info_dict.values():
        serial = entry.serial_number
        if serial in disk_info:
            entry.disk = disk_info[serial]


def _get_address(udev_device):
    """Gets the serial port symlink under '/dev/serial/by-id', mounted drives, or adb serial."""
    devlinks_str = udev_device.get('DEVLINKS', "")
    paths = devlinks_str.split()
    if not devlinks_str:  # Get mounted drives without corresponding symlinks
        if udev_device.get('ID_VENDOR') not in REPEAT_VENDORS:
            return udev_device.get('DEVNAME')

    for path in paths:
        if PERSISTENT_SERIAL_PATH_FOLDER in path:
            return path
        if MOUNTED_DISK_FOLDER in path:
            match = re.match(ADB_SERIAL_REGEX, path)
            if match:
                return match.group(1)
            return path

    return ''


def _get_cambrionix_model(udev_device):
    """Gets the model of the cambrionix."""
    parent = udev_device.parent.parent.parent.parent
    return parent.get("ID_MODEL")


def _get_cambrionix_port_number(dev_path, port_mapping):
    """Get port from dev path.

    Args:
        dev_path (str): udev device dev path.
        port_mapping (dict): dict of port identifiers to port numbers.
    Returns:
       int: port number or None if not found
    Note:
        Port identifier is the last two digits before the ':'
        So for "/devices/pci0000:00/0000:00:1a.0/usb1/"
               "1-1/1-1.5/1-1.5.4/1-1.5.4.3/1-1.5.4.3:1.3/"
               "ttyUSB42/tty/ttyUSB42"
        "1-1.5.4.3:1.3" is the relevant number.
        '1.3' indicates the FTDI interface
        '4.3' indicates the Cambrionix port
        1 - 1.5 indicates the parent usb.
    """
    match = re.search(IDENTIFIER_REGEX, dev_path)
    if match:
        port_identifier = match.group(1)
        if port_identifier in port_mapping:
            return port_mapping[port_identifier]
    return None


def _get_child_addresses(udev_device):
    """Get the addresses for the relevant children of the udev device."""
    all_children = _get_cambrionix_udev_children(udev_device)
    relevant_children = [device for device in all_children
                         if device.get('DEVLINKS') and _get_address(device)]
    if udev_device in relevant_children:
        relevant_children.remove(udev_device)
    grand_children = []
    for child_device in relevant_children:  # Remove children of other cambrionixes
        if _get_product_name(child_device) == usb_config.CAMBRIONIX_NAME:
            grand_children += _get_cambrionix_udev_children(child_device)
            grand_children.remove(child_device)
    return [_get_address(child) for child in relevant_children if child not in grand_children]


def _get_disk_info():
    """Returns a dictionary of disk paths to serial numbers."""
    devices = get_pyudev_list_of_devices(subsystem='block')
    mounted_info = {device.get('ID_SERIAL_SHORT', ""): device.get('DEVNAME')
                    for device in devices}
    return mounted_info


def _get_product_name(udev_device):
    """Gets the product name in a system agnostic way."""
    model = udev_device.get('ID_MODEL', '').replace("_", " ")
    if model == ANDROID_DEVICE_PRODUCT_NAME:  # Actually an Android device.
        model = "Android"
    return model


def _process_udev_device(udev_device):
    """Convert pyudev dict to the system agnostic form."""
    model = None
    dev_path = udev_device.get('DEVPATH', "")

    entry = usb_config.UsbInfo()
    address = _get_address(udev_device)
    if not address:
        return '', {}, None, None
    entry.address = address
    entry.product_id = udev_device.get('ID_MODEL_ID', "")
    entry.vendor_id = udev_device.get('ID_VENDOR_ID', "")
    entry.ftdi_interface = int(udev_device.get('ID_USB_INTERFACE_NUM', 0))
    entry.manufacturer = udev_device.get('ID_VENDOR', "")
    entry.serial_number = udev_device.get('ID_SERIAL_SHORT', "")
    entry.product_name = _get_product_name(udev_device)
    if entry.product_name == usb_config.CAMBRIONIX_NAME:
        entry.child_addresses = _get_child_addresses(udev_device)
        model = _get_cambrionix_model(udev_device)
    return address, entry, dev_path, model


def _get_cambrionix_udev_children(udev_device):
    """Gets all the udev devices representing a cambrionix's children."""
    return list(udev_device.parent.parent.parent.parent.parent.children)
